package incarceration;

/*$$
 * Copyright (c) 1999, Trustees of the University of Chicago
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with 
 * or without modification, are permitted provided that the following 
 * conditions are met:
 *
 *	 Redistributions of source code must retain the above copyright notice,
 *	 this list of conditions and the following disclaimer.
 *
 *	 Redistributions in binary form must reproduce the above copyright notice,
 *	 this list of conditions and the following disclaimer in the documentation
 *	 and/or other materials provided with the distribution.
 *
 * Neither the name of the University of Chicago nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE TRUSTEES OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *$$*/

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Queue;
import java.util.Vector;

//RLR: don't seem to use this (maybe eclipse does...behind the scenes...?)
//import com.sun.xml.internal.messaging.saaj.packaging.mime.util.QEncoderStream;

import uchicago.src.sim.gui.RectNetworkItem;
import uchicago.src.sim.network.DefaultDrawableNode;
import uchicago.src.sim.network.DefaultNode;
import uchicago.src.sim.network.Edge;
import uchicago.src.sim.util.Random;

/**
 * The agent class for the JinGirNew simulation. JinGirNode extends
 * DefaultDrawableNode and so it is both a DefaultNode and a
 * DrawableNonGridNode. All of the important behavior here as to do
 * with creating new links.
 *
 * @version $Revision: 1.1 $ $Date: 2005/08/12 20:04:52 $
 *
 */
@SuppressWarnings("unused")
public class Agent extends DefaultDrawableNode {
	
	private double baseQuality = 0; 
    public int endorsements = 0;
	private double aspirationLevel = 0;

	public int id = -1;
	public int age = -1; 
    public String gender = "female";
    public boolean incarcerated = false;
    public boolean prevIncarceration = false;
    public double probIncarcerationPartnership = 0;  //probability that an agent
    													//that is incarcerated forms a new partnership
    public double incarcerationPenalty = 0; // decrease in quality for being incarcerated
	public int incarcerationTime = 0; // how long an individual will remain incarcerated
	public int incarcerationStart = 0; // time step when the agent was first incarcerated
    public double partnerQuality;  //the average quality of the current partners
    public double potentialPartnerQuality; //the average quality of the current dates
    public int maxPartners = 0;
    public int maxNumberDating = 0;
	private int waitingTime = 0;
	private int waitingThreshold = 0;	
	private int courtShipDuration = 0;
	public double probRecognizeConcurrentTrue = 0;
	public double probRecognizeConcurrentFalse = 0; 
	public double concurrentPenalty = 0;

	
	//in number of months 
	public int avgMonthsDating; //average length of the current dating relationships

	public int timeToFirstPartner = 1;
	public boolean everConcurrent = false;
	public HashMap<Double, Boolean> concurrency = new HashMap<Double, Boolean>();
	public boolean concurrentInPastYear = false;
 	private boolean isMarried = false;	
 	
    public ArrayList<Agent> partners = new ArrayList<Agent>();
    	//individuals with whom the agent is sexually active
    public ArrayList<Couple> currentCouples = new ArrayList<Couple>();
    public ArrayList<Agent> lifetimePartners = new ArrayList<Agent>();
    	// total partners over the lifetime
    public ArrayList<Agent> potentialPartners = new ArrayList<Agent>();  
    	// individuals that the agent is dating
	private ArrayList<Agent> proposers = new ArrayList<Agent>();
		// individuals to whom the agent has proposed dating
	public Agent worsePartner = null;
		//partner to be dropped in favor of new one
	
	public HashMap<Double, ArrayList> stepPartners = new HashMap<Double, ArrayList>();
    public Set<Agent> partnersInLastYear = new HashSet<Agent>();

	private Model model;
	
	public Agent() {}

	public Agent(int x, int y) {
		init(x, y);
	}
	
	public Agent (Model _model, int _id, int _x, int _y) {
		init(_x,_y);
		this.model = _model;
		this.id = _id;
	}
	
    // update quality
        public void updateQuality() {
		    if (incarcerated == true){
		    	double lastQuality = baseQuality;
		    	baseQuality -= (lastQuality * incarcerationPenalty);
		    	if (model.getRDebug() > 1){
		    		System.out.printf( "Quality decrease: Agent %03d from %.3f to %.3f \n", getID(), lastQuality, baseQuality);
		    	}
		    }
	}

    // calculate endorsements
       public int getEndorsements() { return endorsements; }
       public void setEndorsements( Agent a ) {
	   //
       }

    // determines whether the agent should send an offer to a potential partner

	public boolean shouldSendOffer(Agent potentialDate) {
				
		if (maxPartners != 0){
			Double shouldSend = Math.min(((double)avgMonthsDating/courtShipDuration), 1);
			if (potentialDate.getPartners().size() > 0){
				if (model.getRDebug() > 1){
					System.out.printf("\n # of partners: %3d", potentialDate.getPartners().size());
				}
				if ( Math.random() < probRecognizeConcurrentTrue ){
					shouldSend = shouldSend * ( 1 - concurrentPenalty );
					if (model.getRDebug() > 1){
						System.out.printf("\n Recognized true concurrency");
					}
				}
			}
			else
				if ( Math.random() < probRecognizeConcurrentFalse ) {
					shouldSend = shouldSend * ( 1 - concurrentPenalty );
					if (model.getRDebug() > 1){
						System.out.printf("\n # of partners: %3d", potentialDate.getPartners().size());
						System.out.printf("\n Penalized false concurrency");
					}
				}
			if (potentialDate.getIncarcerated() == true){
			// if the individual is in prison then the proposal is evaluated 20% of the time
				if (Math.random() < probIncarcerationPartnership){
						if (potentialPartnerQuality > potentialDate.getBaseQuality()*(1 - shouldSend)){
							return false;
						}
						if (this.aspirationLevel <= potentialDate.getBaseQuality()){
							return false;
						}
						else 
							return true;
				}
				else
					return false;
			}
			if (potentialPartnerQuality > potentialDate.getBaseQuality()* (1 - shouldSend)){
				if (model.getRDebug() > 1){
					// System.out.printf("\n Average potential partner quality: %3f", potentialPartnerQuality);
					// System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					// System.out.printf("\n Weighted quality of potential date: %3f", (potentialDate.getBaseQuality()*(1 - shouldSend)));
					if (shouldSend > 1){
						System.out.printf("\n StopT: %3d", model.getStopT());
						System.out.printf("\n Average months dating: %3d", avgMonthsDating);
						System.out.printf("\n Weight: %3f", (1 - shouldSend));
					}
					// System.out.printf("\n Don't send proposal");
				}
				return false;
			}
			if (this.aspirationLevel >= potentialDate.getBaseQuality()){
				if (model.getRDebug() > 1){
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					System.out.printf("\n Don't send proposal");
				}
				return false;
			}
			else 
				if (model.getRDebug() > 1){
					System.out.printf("\n Average quality of current dates: %3f", potentialPartnerQuality);
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					System.out.printf("\n Send proposal");
				}
			    return true;
		}
		else
			return false;
	}
	
    //gets offer
	public boolean receiveOffer(Agent proposer) {
		
		if (maxPartners != 0){
			Double shouldReceive = Math.min(((double)avgMonthsDating/courtShipDuration), 1);
			if (proposer.getPartners().size() > 0){
				if (model.getRDebug() > 1){
					System.out.printf("\n # of partners: %3d", proposer.getPartners().size());
				}
				if ( Math.random() < probRecognizeConcurrentTrue ){
					shouldReceive = shouldReceive * ( 1 - concurrentPenalty );
					if (model.getRDebug() > 1){
						System.out.printf("\n Recognized true concurrency");
					}
				}
			}
			else
				if ( Math.random() < probRecognizeConcurrentFalse ) {
					shouldReceive = shouldReceive * ( 1 - concurrentPenalty );
					if (model.getRDebug() > 1){
						System.out.printf("\n # of partners: %3d", proposer.getPartners().size());
						System.out.printf("\n Penalized false concurrency");
					}
				}
			if (proposer.getIncarcerated() == true){
				// if the individual is in prison then the proposal is evaluated 20% of the time
				if (Math.random() < probIncarcerationPartnership){
						if (potentialPartnerQuality > proposer.getBaseQuality()* (1 - shouldReceive)){
							return false;
						}
						if (this.aspirationLevel <= proposer.getBaseQuality()){
							return false;
						}
						else 
							return true;
				}
				else
					return false;
			}
			if (potentialPartnerQuality > proposer.getBaseQuality()* (1- shouldReceive )){
				if (model.getRDebug() > 1){
					// System.out.printf("\n Average potential partner quality: %3f", potentialPartnerQuality);
					// System.out.printf("\n Quality of proposer: %3f", proposer.getBaseQuality());
					if (shouldReceive > 1){
						System.out.printf("\n StopT: %3d", model.getStopT());
						System.out.printf("\n Weight: %3f", (1 - shouldReceive));
						System.out.printf("\n Average months dating: %3d", avgMonthsDating);
					}
					//System.out.printf("\n Weighted quality of proposer: %3f", (proposer.getBaseQuality()*(1 - shouldReceive)));
					// System.out.printf("\n Don't accept proposal");
				}
				return false;
			}
			if (this.aspirationLevel >= proposer.getBaseQuality()){
				if (model.getRDebug() > 1){
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of proposer: %3f", proposer.getBaseQuality());
					System.out.printf("\n Don't accept proposal");
				}
				return false;
			}
			else 
				if (model.getRDebug() > 1){
					System.out.printf("\n Average quality of current dates: %3f", potentialPartnerQuality);
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of proposer: %3f", proposer.getBaseQuality());
					System.out.printf("\n Accept proposal");
				}
			    return true;
		}
		else
			return false;
	}
	
	public boolean shouldSendOffer1(Agent potentialDate) {
		
		if (maxPartners != 0){
			Double shouldSend = (double)avgMonthsDating/model.getStopT();
			if (potentialDate.getPartners().size() > 0){
				if (model.getRDebug() > 1){
					System.out.printf("\n # of partners: %3d", potentialDate.getPartners().size());
				}
				if ( Math.random() < probRecognizeConcurrentTrue ){
					shouldSend = shouldSend * ( 1 - concurrentPenalty );
					if (model.getRDebug() > 1){
						System.out.printf("\n Recognized true concurrency");
					}
				}
			}
			else
				if ( Math.random() < probRecognizeConcurrentFalse ) {
					shouldSend = shouldSend * ( 1 - concurrentPenalty );
					if (model.getRDebug() > 1){
						System.out.printf("\n # of partners: %3d", potentialDate.getPartners().size());
						System.out.printf("\n Penalized false concurrency");
					}
				}
			if (potentialDate.getIncarcerated() == true){
			// if the individual is in prison then the proposal is evaluated 20% of the time
				if (Math.random() < probIncarcerationPartnership){
						if (potentialPartnerQuality > potentialDate.getBaseQuality()*(1 - shouldSend)){
							return false;
						}
						if (this.aspirationLevel <= potentialDate.getBaseQuality()){
							return false;
						}
						else 
							return true;
				}
				else
					return false;
			}
			if (potentialPartnerQuality > potentialDate.getBaseQuality()* (1 - shouldSend)){
				if (model.getRDebug() > 1){
					// System.out.printf("\n Average potential partner quality: %3f", potentialPartnerQuality);
					// System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					// System.out.printf("\n Weighted quality of potential date: %3f", (potentialDate.getBaseQuality()*(1 - shouldSend)));
					if (shouldSend > 1){
						System.out.printf("\n StopT: %3d", model.getStopT());
						System.out.printf("\n Average months dating: %3d", avgMonthsDating);
						System.out.printf("\n Weight: %3f", (1 - shouldSend));
					}
					// System.out.printf("\n Don't send proposal");
				}
				return false;
			}
			if (this.aspirationLevel >= potentialDate.getBaseQuality()){
				if (model.getRDebug() > 1){
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					System.out.printf("\n Don't send proposal");
				}
				return false;
			}
			else 
				if (model.getRDebug() > 1){
					System.out.printf("\n Average quality of current dates: %3f", potentialPartnerQuality);
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					System.out.printf("\n Send proposal");
				}
			    return true;
		}
		else
			return false;
	}
	
    //gets offer
	public boolean receiveOffer1(Agent proposer) {
		
		if (maxPartners != 0){
			Double shouldReceive = (double)avgMonthsDating/model.getStopT();
			if (proposer.getPartners().size() > 0){
				if (model.getRDebug() > 1){
					System.out.printf("\n # of partners: %3d", proposer.getPartners().size());
				}
				if ( Math.random() < probRecognizeConcurrentTrue ){
					shouldReceive = shouldReceive * ( 1 - concurrentPenalty );
					if (model.getRDebug() > 1){
						System.out.printf("\n Recognized true concurrency");
					}
				}
			}
			else
				if ( Math.random() < probRecognizeConcurrentFalse ) {
					shouldReceive = shouldReceive * ( 1 - concurrentPenalty );
					if (model.getRDebug() > 1){
						System.out.printf("\n # of partners: %3d", proposer.getPartners().size());
						System.out.printf("\n Penalized false concurrency");
					}
				}
			if (proposer.getIncarcerated() == true){
				// if the individual is in prison then the proposal is evaluated 20% of the time
				if (Math.random() < probIncarcerationPartnership){
						if (potentialPartnerQuality > proposer.getBaseQuality()* (1 - shouldReceive)){
							return false;
						}
						if (this.aspirationLevel <= proposer.getBaseQuality()){
							return false;
						}
						else 
							return true;
				}
				else
					return false;
			}
			if (potentialPartnerQuality > proposer.getBaseQuality()* (1- shouldReceive )){
				if (model.getRDebug() > 1){
					// System.out.printf("\n Average potential partner quality: %3f", potentialPartnerQuality);
					// System.out.printf("\n Quality of proposer: %3f", proposer.getBaseQuality());
					if (shouldReceive > 1){
						System.out.printf("\n StopT: %3d", model.getStopT());
						System.out.printf("\n Weight: %3f", (1 - shouldReceive));
						System.out.printf("\n Average months dating: %3d", avgMonthsDating);
					}
					//System.out.printf("\n Weighted quality of proposer: %3f", (proposer.getBaseQuality()*(1 - shouldReceive)));
					// System.out.printf("\n Don't accept proposal");
				}
				return false;
			}
			if (this.aspirationLevel >= proposer.getBaseQuality()){
				if (model.getRDebug() > 1){
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of proposer: %3f", proposer.getBaseQuality());
					System.out.printf("\n Don't accept proposal");
				}
				return false;
			}
			else 
				if (model.getRDebug() > 1){
					System.out.printf("\n Average quality of current dates: %3f", potentialPartnerQuality);
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of proposer: %3f", proposer.getBaseQuality());
					System.out.printf("\n Accept proposal");
				}
			    return true;
		}
		else
			return false;
	}
	
    // determines whether the agent should send an offer to a potential partner

	public boolean shouldSendOfferUnweighted(Agent potentialDate) {
				
		if (maxPartners != 0){
			if (potentialDate.getIncarcerated() == true){
			// if the individual is in prison then the proposal is evaluated 20% of the time
				if (Math.random() < probIncarcerationPartnership){
						if (potentialPartnerQuality > potentialDate.getBaseQuality()){
							return false;
						}
						if (this.aspirationLevel <= potentialDate.getBaseQuality()){
							return false;
						}
						else 
							return true;
				}
				else
					return false;
			}
			if (potentialPartnerQuality > potentialDate.getBaseQuality()){
				if (model.getRDebug() > 1){
					System.out.printf("\n Average potential partner quality: %3f", potentialPartnerQuality);
					System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					System.out.printf("\n Don't send proposal");
				}
				return false;
			}
			if (this.aspirationLevel >= potentialDate.getBaseQuality()){
				if (model.getRDebug() > 1){
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					System.out.printf("\n Don't send proposal");
				}
				return false;
			}
			else 
				if (model.getRDebug() > 1){
					System.out.printf("\n Average quality of current dates: %3f", potentialPartnerQuality);
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					System.out.printf("\n Send proposal");
				}
			    return true;
		}
		else
			return false;
	}
	
    //gets offer
	public boolean receiveOfferUnweighted(Agent proposer) {
		
		if (maxPartners != 0){
			if (proposer.getIncarcerated() == true){
				// if the individual is in prison then the proposal is evaluated 20% of the time
				if (Math.random() < probIncarcerationPartnership){
						if (potentialPartnerQuality > proposer.getBaseQuality()){
							return false;
						}
						if (this.aspirationLevel <= proposer.getBaseQuality()){
							return false;
						}
						else 
							return true;
				}
				else
					return false;
			}
			if (potentialPartnerQuality > proposer.getBaseQuality()){
				if (model.getRDebug() > 1){
					System.out.printf("\n Average potential partner quality: %3f", potentialPartnerQuality);
					System.out.printf("\n Quality of proposer: %3f", proposer.getBaseQuality());
					System.out.printf("\n Don't accept proposal");
				}
				return false;
			}
			if (this.aspirationLevel >= proposer.getBaseQuality()){
				if (model.getRDebug() > 1){
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of proposer: %3f", proposer.getBaseQuality());
					System.out.printf("\n Don't accept proposal");
				}
				return false;
			}
			else 
				if (model.getRDebug() > 1){
					System.out.printf("\n Average quality of current dates: %3f", potentialPartnerQuality);
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of proposer: %3f", proposer.getBaseQuality());
					System.out.printf("\n Accept proposal");
				}
			    return true;
		}
		else
			return false;
	}
	
	public boolean shouldSendOfferWeighted(Agent potentialDate) {
		
		if (maxPartners != 0){
			if (potentialDate.getIncarcerated() == true){
			// if the individual is in prison then the proposal is evaluated 20% of the time
				if (Math.random() < probIncarcerationPartnership){
					if (courtShipDuration != 0){
						if (avgMonthsDating > courtShipDuration){
							return false;
						}
						else {
							Double shouldSend = (potentialDate.getBaseQuality() * 
							  (Math.abs((this.courtShipDuration - this.avgMonthsDating)/this.courtShipDuration)) *
							((maxPartners - partners.size())/maxPartners));
							if (potentialPartnerQuality > shouldSend){
							return false;
							}
						}
					}
					else {
						if (potentialPartnerQuality > (potentialDate.getBaseQuality() * 
							((maxPartners - partners.size())/maxPartners))){
							return false;
						}
					}
					if (this.aspirationLevel <= potentialDate.getBaseQuality()){
						return false;
					}
					else 
						return true;
				}
				else
					return false;
			}
			if (courtShipDuration != 0){
				if (avgMonthsDating > courtShipDuration){
					if (model.getRDebug() > 1){
						System.out.printf("\n Average months dating: %3d", avgMonthsDating);
						System.out.printf("\n Courtship duration: %3d", courtShipDuration);
						System.out.printf("\n Don't send proposal");
					}
					return false;
				}
				else {
					Double shouldSend = (potentialDate.getBaseQuality() * 
					  (Math.abs((this.courtShipDuration - this.avgMonthsDating)/this.courtShipDuration)) *
					  ((maxPartners - partners.size())/maxPartners));
					if (potentialPartnerQuality > shouldSend){
						if (model.getRDebug() > 1){
							System.out.printf("\n Average quality of current dates: %3f", potentialPartnerQuality);
							System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
							System.out.printf("\n Weighted quality of potential date: %3f", shouldSend);
							System.out.printf("\n Don't send proposal");
						}
						return false;
					}
				}
			}
			else {
				if (potentialPartnerQuality > (potentialDate.getBaseQuality() * 
						((maxPartners - partners.size())/maxPartners))){
					if (model.getRDebug() > 1){
						System.out.printf("\n Average quality of current dates: %3f", potentialPartnerQuality);
						System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
						System.out.printf("\n Weighted quality of potential date: %3f", (potentialDate.getBaseQuality() * 
								((maxPartners - partners.size())/maxPartners)));
						System.out.printf("\n Don't send proposal");
					}
					return false;
				}
			}
			
			if (this.aspirationLevel >= potentialDate.getBaseQuality()){
				if (model.getRDebug() > 1){
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					System.out.printf("\n Don't send proposal");
				}
				return false;
			}
			else 
				if (model.getRDebug() > 1){
					System.out.printf("\n Average quality of current dates: %3f", potentialPartnerQuality);
					System.out.printf("\n Aspiration level: %3f", aspirationLevel);
					System.out.printf("\n Quality of potential date: %3f", potentialDate.getBaseQuality());
					System.out.printf("\n Send proposal");
				}
			    return true;
		}
		else
			return false;
	}
	
    //gets offer
	public boolean receiveOfferWeighted(Agent proposer) {
		
		if (maxPartners != 0){
			if (proposer.getIncarcerated() == true){
				return Math.random() < probIncarcerationPartnership ? true : false;
				// if the individual is in prison then the proposal goes through 20% of the time
			}
			if (courtShipDuration != 0){
				if (avgMonthsDating > courtShipDuration){
					return false;
				}
				else {
					Double shouldReceive = (proposer.getBaseQuality() * 
					  (Math.abs((this.courtShipDuration - this.avgMonthsDating)/this.courtShipDuration)) *
					  ((maxPartners - partners.size())/maxPartners));
					if (potentialPartnerQuality > shouldReceive){
						return false;
					}
				}
			}
			else if (potentialPartnerQuality > (proposer.getBaseQuality() * 
					(Math.abs((maxPartners - partners.size())/maxPartners)))){
				return false;
			}
			
			if (this.aspirationLevel <= proposer.getBaseQuality()){
				return false;
			}
			else 
			    return true;
		}
		else
			return false;
	}
	
	// compare dating partner with current partners
	public Agent comparePartners(Agent potentialPartner, ArrayList<Agent> list){
		//if (model.getRDebug() > 0){
		//	System.out.printf("Starting to compare partners in Agent.java");
		//}
		worsePartner = null;
		for (Agent partner: list){
			Double shouldDump = 1.0;
			if (partner.getPartners().size() > 1){
				if (model.getRDebug() > 1){
					System.out.printf("\n # of partners: %3d", partner.getPartners().size());
				}
				if ( Math.random() < probRecognizeConcurrentTrue ){
					shouldDump = shouldDump * ( 1 - concurrentPenalty );
					if (model.getRDebug() > 1){
						System.out.printf("\n Recognized true concurrency");
					}
				}
			}
			else
				if ( Math.random() < probRecognizeConcurrentFalse ) {
					shouldDump = shouldDump * ( 1 - concurrentPenalty );
					if (model.getRDebug() > 1){
						System.out.printf("\n # of partners: %3d", partner.getPartners().size());
						System.out.printf("\n Penalized false concurrency");
					}
				}
			if (potentialPartner.getBaseQuality() > partner.getBaseQuality() * (shouldDump)){
				if (model.getRDebug() > 1){
					System.out.printf("%3d to be replaced", partner.getID());
				}
				worsePartner = partner;
				break;
			}
		}
		return worsePartner;
	}

    // end particular courtship
	public void courtshipBroken(Agent partner) {
	        potentialPartners.remove(partner);
	        partners.remove(partner);
	        waitingTime = 0;
	}
	
    //updates the agents
	public void update() {
		if (partners.size() < maxPartners) {
			waitingTime++;
			if (this.waitingTime > this.waitingThreshold) {
				if (model.getToleranceChange() == false ) {
					decreaseAspiration();	
				}
				if (model.getToleranceChange() == true) {
					decreaseAspiration();
					increaseConcurrencyTolerance();
				}
			}
		}
		
		calcRelationshipStats();
		
		if (this.aspirationLevel <= partnerQuality) {
			increaseAspiration();
		}
		updateQuality();
	}
	
	public void updatePartnerList() {
		stepPartners.put(model.getTickCount(), partners);
		if (stepPartners.size() > 1){
			concurrency.put(model.getTickCount(), true);
		}
		if (stepPartners.size() < 2){
			concurrency.put(model.getTickCount(), false);
		}
		if (model.getTickCount() >= 52){
			for (double i = (model.getTickCount() - 52); i <= model.getTickCount(); i ++){
			    if (stepPartners.containsKey(i)) {
				partnersInLastYear.addAll(stepPartners.get(i));
			    }
			}
		}
		if (model.getTickCount() >= 52){
			for (double i = (model.getTickCount() - 52); i <= model.getTickCount(); i ++){
			    if (concurrency.containsKey(i) && concurrency.get(i) == true) {
				concurrentInPastYear = true;
			    }
			}
		}
	}
      
    // increase or decrease aspiration in response to too long a waiting time or broken courtship
	public void decreaseAspiration() {
		this.aspirationLevel -= 0.1;
	}
	
	public void increaseAspiration() {
	        this.aspirationLevel += 0.1;
	}
	
	public void increaseTimeToFirstPartner() {
		this.timeToFirstPartner += 1;
	}
	
	public void increaseConcurrencyTolerance() {
		if (model.getRDebug() > 2){
			System.out.printf("False: %.2f \n", probRecognizeConcurrentFalse);
			System.out.printf("True: %.2f \n", probRecognizeConcurrentTrue);
			System.out.printf("Penalty: %.2f \n", concurrentPenalty);
		}
		this.probRecognizeConcurrentFalse = Math.max(probRecognizeConcurrentFalse - 0.01, 0);
		this.probRecognizeConcurrentTrue = Math.max((probRecognizeConcurrentTrue - 0.01), 0);
		this.concurrentPenalty = Math.max(concurrentPenalty - 0.01, 0);
		if (model.getRDebug() > 2){
			System.out.printf("False: %.2f \n", probRecognizeConcurrentFalse);
			System.out.printf("True: %.2f \n", probRecognizeConcurrentTrue);
			System.out.printf("Penalty: %.2f \n", concurrentPenalty);
		}
	}

	//calcRelationshipStats()
	//strings together all of the different statistics necessary
	public void calcRelationshipStats(){
		calcPartnerQuality();
		calcPotentialPartnerQuality();
		calcAvgMonthsDating();
	}
    //calcPartnerQuality()
    //calculates the average quality over all of the partners weighting
        public void calcPartnerQuality(){
	    partnerQuality = 0;
	    for (Agent partner: partners ){
	    	partnerQuality += partner.getBaseQuality();
	    }
	    if (partners.size() > 1) {
	    	partnerQuality /= partners.size();
	    }
	}
     //calcPotentialPartnerQuality()
     //calculates the average quality over all of the partners weighting
        public void calcPotentialPartnerQuality(){
    	potentialPartnerQuality = 0;
    	    for (Agent potentialPartner: potentialPartners ){
    		potentialPartnerQuality += potentialPartner.getBaseQuality();
    	    }
    	    if (potentialPartners.size() > 1) {
    	    	potentialPartnerQuality /= potentialPartners.size();
    	    }
    	    else
    	    	potentialPartnerQuality = 0;
    	}
        
     //calcAvgMonthsDating()
     //calculates the average duration of all of the dating relationships
     public void calcAvgMonthsDating(){
       	avgMonthsDating = 0;
       		if (currentCouples.size() > 0){
       			for (Couple couple: currentCouples ){
       				avgMonthsDating += couple.getNumMonthsDating();
       			}
       			avgMonthsDating /= currentCouples.size();
       		}
       	    else
       	    	avgMonthsDating = 0;
       	}

	/**
	 * Initialize this Agent, this takes the place of the
	 * constructor when a Agent is created from a file using its
	 * no-arg constructor.
	 */
	public void init(int x, int y) {
		RectNetworkItem rect = new RectNetworkItem(x, y);
		setDrawable(rect);
	}

	/**
	 * Makes an edge to the specified node and from the specifed node to
	 * thisAgent if both nodes do not already have edges to each
	 * other and if adding the edge keeps their degrees less than
	 * maxDegree. The edges is displayed in the specified color.
	 */
	public void makeEdgeToFrom(DefaultNode node, int maxDegree, Color color) {
		if ((! hasEdgeTo(node)) && getOutDegree() < maxDegree &&
				node.getOutDegree() < maxDegree) {
			Edge edge = new EdgeLink(this, node, color);
			addOutEdge(edge);
			node.addInEdge(edge);
			Edge otherEdge = new EdgeLink(node, this, color);
			node.addOutEdge(otherEdge);
			addInEdge(otherEdge);
		}
	}

	/**
	 * Creates a new edge between this Agent and a node chosen
	 * at random from the specified list. This edge is created via
	 * <tt>makeEdgeFromTo</tt> and so those conditions must be met as
	 * well.
	 */
	public void meetRandom(ArrayList list, int maxDegree) {
		int index = Random.uniform.nextIntFromTo(0, list.size() - 1);
		Agent node = (Agent)list.get(index);
		while (this.equals(node)) {
			index = Random.uniform.nextIntFromTo(0, list.size() - 1);
			node = (Agent)list.get(index);
		}

		makeEdgeToFrom(node, maxDegree, Color.red);

		if (model.rDebug > 2)
		    System.out.printf("-Random edge: %03d and %03d \n", id, node.getID() );
	}

	/**
	 * Creates an edge between two nodes that both link to this
	 * Agent.  This edge is created via <tt>makeEdgeFromTo</tt>
	 * and so those conditions must be met as well.
	 */
	public void meetNeighbor(int maxDegree) {
		if (getOutDegree() > 1) {
			Agent jNode = (Agent)getRandomNodeOut();
			Agent kNode = (Agent)getRandomNodeOut();

			while (jNode.equals(kNode)) {
				kNode = (Agent)getRandomNodeOut();
			}
			jNode.makeEdgeToFrom(kNode, maxDegree, Color.green);

		if (model.rDebug > 2)
		    System.out.printf("-Neighbor edge: %03d and %03d \n", 
				      jNode.getId(), kNode.getID() );
		}
	}

	/**
	 * Removes a link between this Agent and one chosen at
	 * random from those linked to this Agent.
	 */
	public void removeFriend() {
		Agent jNode = (Agent)getRandomNodeOut();
		// will be null if no outEdges.
		if (jNode != null) {
			removeEdgesTo(jNode);
			jNode.removeEdgesFrom(this);
			removeEdgesFrom(jNode);
			jNode.removeEdgesTo(this);
		}
	}

	public int getAge() { return age; }
	public void setAge(int age) { this.age = age; }

	public double getAspirationLevel() { return aspirationLevel; }
	public void setAspirationLevel(double aspiration) { this.aspirationLevel = aspiration; 	}

	public double getBaseQuality() { return baseQuality; }
	public void setBaseQuality(double baseQuality) { this.baseQuality = baseQuality; }

	public int getCourtShipTime() { return waitingThreshold; }
	public void setCourtShipTime(int courtShipTime) { this.waitingThreshold = courtShipTime; }

	public String getGender() {return gender; }
	public void setGender(String gender) { this.gender = gender; }

    public int getID() { return id; }
    
    public int getTimeToFirstPartner() { return timeToFirstPartner; }

	public int getWaitingThreshold() { return waitingThreshold; }

	public void setWaitingThreshold(int waitingThreshold) { this.waitingThreshold = waitingThreshold; }
	public int getWaitingTime() { return waitingTime; }

	public void setWaitingTime(int waitingTime) {
		this.waitingTime = waitingTime;
	}

    public int getMaxPartners() { return maxPartners; }
	public void setMaxPartners( int p ) { maxPartners = p; }
	
	public int getMaxNumberDating() { return maxNumberDating; }
	public void setMaxNumberDating( int m ) { maxNumberDating = m; }

	public int getCourtShipDuration() {
		return courtShipDuration;
	}

	public void setCourtShipDuration(int courtShipDuration) {
		this.courtShipDuration = courtShipDuration;
	}

	public boolean isMarried() {
		return isMarried;
	}

	public void setMarried(boolean isMarried) {
		this.isMarried = isMarried;
	}

	public ArrayList<Agent> getProposers() {
		return proposers;
	}

	public void setProposers(ArrayList<Agent> proposers) {
		this.proposers = proposers;
	}

	public ArrayList getPotentialPartners() { return potentialPartners; }
	public void setPotentialPartners(ArrayList<Agent> p ) { potentialPartners = p; 	}

    public ArrayList getPartners() { return partners; }
    public void setPartners( ArrayList<Agent> p ) { partners = p; }

    public boolean getIncarcerated() { return incarcerated; }
    public void setIncarcerated( boolean b ) { incarcerated = b; }

    public boolean getPrevIncarceration() { return prevIncarceration; }
    public void setPrevIncarceration( boolean p ) { prevIncarceration = p; }
        
    public boolean getEverConcurrent() {return everConcurrent; }
    public void setEverConcurrent ( boolean e ) { everConcurrent = e; }
    
    public int getIncarcerationTime() { return incarcerationTime; }
    public void setIncarcerationTime( int t ) { incarcerationTime = t; }
    
    public int getIncarcerationStart() { return incarcerationStart; }
    public void setIncarcerationStart( int s ) { incarcerationStart = s; }
    
    public ArrayList getLifetimePartners() {return lifetimePartners; }
    
    public double getProbRecognizeConcurrentTrue() { return probRecognizeConcurrentTrue; }
    public void setProbRecognizeConcurrentTrue( double r ) { probRecognizeConcurrentTrue = r; }
    
    public double getProbRecognizeConcurrentFalse() { return probRecognizeConcurrentFalse; }
    public void setProbRecognizeConcurrentFalse( double r ) { probRecognizeConcurrentFalse = r; }
    
    public double getConcurrentPenalty() { return concurrentPenalty; }
    public void setConcurrentPenalty( double c ) { concurrentPenalty = c; }
    
    public double getProbIncarcerationPartnership() { return probIncarcerationPartnership; }
    public void setProbIncarcerationPartnership( double i ) { probIncarcerationPartnership = i; }

    public double getIncarcerationPenalty() { return incarcerationPenalty; }
    public void setIncarcerationPenalty( double p ) { incarcerationPenalty = p; }

}
