package incarceration;


import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Color;
import java.util.ArrayList;

import uchicago.src.sim.engine.*;
import uchicago.src.sim.util.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.space.Object2DTorus;
import uchicago.src.sim.engine.AbstractGUIController;
import uchicago.src.sim.analysis.*;


public class GUIModel extends Model {
	
    public AbstractGraphLayout graphLayout;
    public Histogram degreeDist;

    public NetSequenceGraph graph;
    public DisplaySurface surface;

    public  OpenSequenceGraph  	graph1;


	/////////////////////////////////////////////////////////////////////
	// setup
	//
	// this runs automatically when the model starts
	// and when you click the reload button, to "tear down" any 
	// existing display objects, and get ready to initialize 
	// them at the start of the next 'run'.
	//
	public void setup() {
		
		super.setup();  // RLR the super class does conceptual-model setup

		//RLR
		// NOTE: you may want to set these next two to 'true'
		// if you are on a windows machine.  that would tell repast
		// to by default send System.out and .err output to
		// a special repast output window.
		AbstractGUIController.CONSOLE_ERR = false;
		AbstractGUIController.CONSOLE_OUT = false;
		AbstractGUIController.UPDATE_PROBES = true;

		// RLR Random.createUniform();

			if (surface != null) surface.dispose();
			if (degreeDist != null) degreeDist.dispose();
			if (graph != null) graph.dispose();

			surface = null;
			schedule = null;
			degreeDist = null;
			graph = null;

			System.gc();

			surface = new DisplaySurface(this, "JinGirModelII Display");
			registerDisplaySurface("Main Display", surface);
			schedule = new Schedule();
			agentList = new ArrayList();
			worldXSize = 500;
			worldYSize = 500;

		if ( rDebug > 0 )
			System.out.printf( "<== GUIModel setup() done.\n" );
	}

	/////////////////////////////////////////////////////////////////////
	// begin
	//
	// this runs when you click the "initialize" button
	// (the button with the single arrow that goes around in a circle)
	//
	public void begin()	{		
		buildModel();
		buildDisplay();
		buildSchedule();
		graphLayout.updateLayout();
		surface.display();
		//	if (showHist) degreeDist.display();
		
		//	if (showPlot) graph.display();
	}

	/////////////////////////////////////////////////////////////////////
	// buildDisplay
	//
	// builds the display and display related things
	//
	public void buildDisplay() {
		if ( rDebug > 0 )
			System.out.printf( "==> GUIModel buildDisplay...\n" );

		if (layoutType.equals("KK")) {
			graphLayout = new KamadaGraphLayout(agentList, worldXSize, worldYSize,
					surface, updateEveryN);
		} else if (layoutType.equals("Fruch")) {
			graphLayout = new FruchGraphLayout(agentList, worldXSize, worldYSize, surface, updateEveryN);
		} else if (layoutType.equals("CircleLayout")) {
			graphLayout = new CircularGraphLayout(agentList, worldXSize,
					worldYSize);
		}

		// these four lines hook up the graph layouts to the stop, pause, and
		// exit buttons on the toolbar. When stop, pause, or exit is clicked
		// the graph layouts will interrupt their layout as soon as possible.
		Controller c = (Controller) getController();
		c.addStopListener(graphLayout);
		c.addPauseListener(graphLayout);
		c.addExitListener(graphLayout);

		Network2DDisplay display = new Network2DDisplay(graphLayout);
		surface.addDisplayableProbeable(display, "JinGirNew Display");

		// add the display as a Zoomable. This means we can "zoom" in on
		// various parts of the network.
		surface.addZoomable(display);
		surface.setBackground(java.awt.Color.white);
		addSimEventListener(surface);
	}

	////////////////////////////////////////////////////////////////
	// buildSchedule
	//
	// This builds the entire schedule, i.e., 
	//  - the base model step (calls stepReport)
	//  - display steps.

	public void buildSchedule() {

		if ( rDebug > 0 )
			System.out.printf( "==> GUIModel buildSchedule...\n" );
		
		/*
			initialAction = schedule.scheduleActionAt(1, this, "initialAction");
			schedule.scheduleActionAt(initialSteps, this, "removeInitialAction",
					Schedule.LAST);
			schedule.scheduleActionBeginning(initialSteps + 1, this, "mainAction");
			schedule.scheduleActionAtEnd(graph, "writeToFile");
		*/

		// schedule the current GUIModel's step() function
		// to execute every time step starting with time step 0
		schedule.scheduleActionBeginning( 0, this, "step" );

		// schedule the current GUIModel's processEndOfRun() 
		// function to execute at the end of the run
		schedule.scheduleActionAtEnd( this, "processEndOfRun" );
	}
	
	/*
	 * Creates a histogram of the degree distribution.
	 */
	private void makeHistogram() {

		degreeDist = new Histogram("Degree Distribution", maxDegree + 1, 0,
				maxDegree + 1, this);

		degreeDist.createHistogramItem("Degree Distribution", agentList,
		"getOutDegree");
	}
	
	/*
	 * Creates a Plot of the Clustering Coefficient, the avg. density,
	 * and the component count.
	 */
	private void makePlot() {

//		graph = new NetSequenceGraph("Network Stats", this, "./net.txt", PlotModel.CSV, agentList);
//		graph.setAxisTitles("Time", "Statistic Value");
//		graph.setXRange(0, 50);
//		graph.setYRange(0, numNodes);
//		graph.plotClusterCoefficient("Cluster Coef.", Color.red, NetSequenceGraph.SQUARE);

		graph = new NetSequenceGraph("Network Stats", this, "./net.txt",
				PlotModel.CSV, this.agentList);
		graph.setAxisTitles("Time", "Statistic Value");
		graph.addSequence("Number of Couples", new Sequence() {
			  public double getSValue() {
				  double distance = returnNumberCouples();
			    return distance;
			  }
			});
		graph.plotClusterCoefficient("Cluster Coef.", Color.red, NetSequenceGraph.SQUARE);
		
	}


	///////////////////////////////////////////////////////////////////////////////
	// step
	//
	// executed each step of the model.
	// Ask the super-class to do its step() method,
	// and then this does display related activities.
	//
	public void step() {

		super.step();			// the model does whatever it does
		

		if (showHist) makeHistogram();
		if (showPlot) makePlot();

		// add things after this for all displays (graphs, etc)
		//RLR graph.step();

	}

	// processEndOfRun
	// called once, at end of run.
	public void processEndOfRun ( ) {
		if ( rDebug > 0 )  
			System.out.printf("\n\n===== GUIModel processEndOfRun =====\n\n" );
		applyAnyStoredChanges();
		endReportFile();
		this.fireStopSim();
	}



/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
//   ****  NO NEEd TO CHANGE THE REST OF THIS  *****

	////////////////////////////////////////////////////////////////////
	// main entry point
	public static void main( String[] args ) {

		uchicago.src.sim.engine.SimInit init =
			new uchicago.src.sim.engine.SimInit();
		GUIModel model = new GUIModel();

		//System.out.printf("==> GUIMOdel main...\n" );

		// set the type of model class, this is necessary
		// so the parameters object knows whether or not
		// to do GUI related updates of panels,etc when a
		// parameter is changed
		model.setModelType("GUIModel");

        // Do this to set the Update Probes option to true in the
        // Repast Actions panel
        Controller.UPDATE_PROBES = true;

		model.setCommandLineArgs( args );
		init.loadModel( model, null, false ); // does setup()

		// this new function calls ProbeUtilities.updateProbePanels() and 
		// ProbeUtilities.updateModelProbePanel()
		model.updateAllProbePanels();

	}

}
