package incarceration;

// A simple model of partnering.

import java.io.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import jsc.datastructures.PairedData;
import jsc.correlation.LinearCorrelation;

import cern.jet.random.Normal;
import cern.jet.random.engine.MersenneTwister;

//RLR import flanagan.math.PsRandom;

import uchicago.src.collection.RangeMap;
import uchicago.src.reflector.ListPropertyDescriptor;
import uchicago.src.sim.analysis.Histogram;
import uchicago.src.sim.analysis.NetSequenceGraph;
import uchicago.src.sim.engine.BasicAction;
import uchicago.src.sim.engine.Schedule;
import uchicago.src.sim.engine.SimModelImpl;
import uchicago.src.sim.util.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.space.*;

enum askingMethod {
	MENASK,
	WOMENASK,
	BOTHASK
};

@SuppressWarnings("unused")
public class Model extends ModelParameters {
	
	// instance variables for run-time parameters
	public int numNodes = 250;
	public int worldXSize = 400;
	public int worldYSize = 400;
	public int updateEveryN = 5;
	public int initialSteps = 1;
	public double removeProb = 0.005;
	public int maxDegree = 10;
	public double Rsub0 = 0.005;
	public double Rsub1 = 2;
    public double sexRatio = 0.5; //proportion of men in the initial population
    public double probRandomPartner = 0.2; 
    //probability that a partner is found outside the social network
    public askingMethod ask = askingMethod.BOTHASK; 
    //men only asking is the default value
	public String askMethod = "both";
	public boolean networkOn = true;
	public boolean weightedSwitch = true;
	public String qualityDistribution = "normal";
	public int degFreedom = 25;
	public boolean eqQuality = true;

	// an iv used by repast for keeping track of model steps
	public Schedule schedule;
	public BasicAction initialAction;
	public RangeMap neighborMap = new RangeMap();
	public RangeMap removalMap = new RangeMap();
	public String layoutType = "Fruch";
    public boolean showHist = true;
    public boolean showPlot = true;
    public int reportFileContents = 0;

	public ArrayList<Agent>agentList = new ArrayList<Agent>();
	public ArrayList<Couple> couples = new ArrayList<Couple>();
    public ArrayList<Couple> married = new ArrayList<Couple>();
	public ArrayList<Agent> females = new ArrayList<Agent>();
	public ArrayList<Agent> males = new ArrayList<Agent>();
	public ArrayList<Agent> neverCoupled = new ArrayList<Agent>();
	ArrayList<Agent> maleFriends = new ArrayList<Agent>();
	ArrayList<Agent> femaleFriends = new ArrayList<Agent>();
	ArrayList<Agent> friends = new ArrayList<Agent>();
	
	public int lastAgentId = 0, lastCoupleId = 0;

    public double meanAspiration = 50;
    public double sdAspiration = 25;
    // gamma distribution parameters: expected value = alpha/lambda
    public double alphaMale = 20;
    public double lambdaMale = 1;
    public double alphaFemale = 2;
    public double lambdaFemale = 1;
    public double meanQuality = 50;
    public double sdQuality = 25;
        //courtship duration - time agents wait to enter into a sexual relationship
    public double meanDuration = 10;
    public double sdDuration = 2;
        //waiting period to start decreasing aspirations
    public int minWaiting = 5;
    public int maxWaiting = 10;
    public int meanNumberDating = 10;
    public int sdNumberDating = 5;
    	//number of agents and individual can date at one time
	public double probRecognizeConcurrentTrue = 0.2;
	public double probRecognizeConcurrentFalse = 0.1; 
	public double concurrentPenalty = 0.6;
	public boolean toleranceChange = true;

       //estimated numbers based on Western and Pettit's (2004) estimate of incarceration rates of black men 25-29
       // divided rates by 5 (to get to yearly) and then by 12 to estimate monthly
       // clearly there are important assumptions here, but they give at least a ballpark

    public double probIncarcerationMale = 0; //0.001;
    public double probIncarcerationFemale = 0; // 0.0001;
    public double meanTime = 16;
    public double sdTime = 4;
    public double probIncarcerationPartnership = 0.2;
    public double probIncarcerationBreakup = 0.5;
    public double incarcerationPenalty = 0.1; //fraction of quality that is lost to incarceration
    public double probRecidivate = 0; //probability of non-random re-incarceration

        //aggregate measures - mostly to determine how things are running
    public double avgQuality;
    public double avgAspiration;
    public double avgNumPartners;
    public double avgNumProposers;
    public double avgNumPotential;
    public double coupleCorr;
    public double avgMonthsDating;
    public double percentMarried;
    public int currentlyIncarcerated;
    public int everIncarcerated;
    public int part1;
    public int part2;
    public int part3;
    public int part4plus;
    public double avgTimeToFirstPartner;
    public double fractionConcurrent;
    public double fractionConcurrentInYear;
    public double avgLifetimePartners;
    public int partLife1;
    public int partLife2;
    public int partLife3;
    public int partLife4plus;
    public int partLife1F;
    public int partLife2F;
    public int partLife3F;
    public int partLife4F;
    public int partLife5F;
    public int partLife6to10F;
    public int partLife11to20F;
    public int partLife21plusF;
    public int partLife1M;
    public int partLife2M;
    public int partLife3M;
    public int partLife4M;
    public int partLife5M;
    public int partLife6to10M;
    public int partLife11to20M;
    public int partLife21plusM;
    public double avgPartnersInLastYear;
	public double avgQualityFemale;
	public double avgQualityMale;
	public double percentSexuallyActiveF;
	public double percentSexuallyActiveM;
	public int currentlyIncarceratedM;
	public int currentlyIncarceratedF;
	public int everIncarceratedF;
	public int everIncarceratedM;
	public double fractionConcurrentM;
	public double fractionConcurrentF;
	public double avgNumPartnersM;
	public double avgNumPartnersF;
	public double avgLifetimePartnersM;
	public double avgLifetimePartnersF;
	public double avgPartnersInLastYearM;
	public double avgPartnersInLastYearF;
    public int part1M;
    public int part2M;
    public int part3M;
    public int part4M;
    public int part5plusM;
    public int part1F;
    public int part2F;
    public int part3F;
    public int part4F;
    public int part5plusF;
	public double fractionConcurrentIM;
	public double avgPartnersInLastYearIM;
	public double avgLifetimePartnersIM;
    
    public String edgeListReportFileName = "";
    public PrintWriter edgeListReportFile, edgeListPlainTextReportFile;
    
	/////////////////////////////////////////////////////////////////////////////
	// To add a new parameter that can be set at run time (via command line or gui)
	// do the following steps:
	// 0. Add an instance variable (field) for the parameter to Model.java .
	// 1. In the method addModelSpecificParameters
	//    add alias and long name for Model parameters you want to set at run time
	//    the long name should be same as instance variable
	// 2. Add setter and getter methods for the parameter
	// 3. In the method getInitParam, add the parameter name to the array of Strings.
	// 4. Compile and test by running model and seeing that the parameter:
	//    - is on the GUI panel, in the report file, and settable on the command line,
	//      eg     bin/batchrun.sh newParName=1.4
	// Note: the generic parameters from ModelParameters are already available.
	//       enter    bin/guirun.sh -h  
	// to see all parameters and defaults, including new ones you add.

	public void addModelSpecificParameters () {
		parametersMap.put( "X", "worldXSize" );
		parametersMap.put( "Y", "worldYSize" );
		parametersMap.put( "nA", "numNodes" );
		parametersMap.put( "N", "updateEveryN" );
		parametersMap.put( "L", "LayoutType" );
		parametersMap.put( "R", "RemoveProb" );
		parametersMap.put( "S", "StartRemoveAfter" );
		parametersMap.put( "max", "MaxDegree" );
		parametersMap.put( "alm", "alphaMale" );
		parametersMap.put( "lam", "lambdaMale" );
		parametersMap.put( "alf", "alphaFemale" );
		parametersMap.put( "laf", "lambdaFemale" );
		parametersMap.put( "Q", "meanQuality");
		parametersMap.put( "sdQ", "sdQuality");
		parametersMap.put( "A", "meanAspiration");
		parametersMap.put( "sdA", "sdAspiration");
		parametersMap.put( "R1", "Rsub1" );
		parametersMap.put( "R0", "Rsub0" );
		parametersMap.put( "dist", "degreeDist" );
		parametersMap.put( "plot", "showPlot" );
		parametersMap.put( "rD", "rDebug" );
		parametersMap.put( "sR", "sexRatio" );
		parametersMap.put( "ask", "askMethod" );
		parametersMap.put( "pR", "probRandomPartner" );
		parametersMap.put( "ELr", "edgeListReportFileName");
		parametersMap.put( "mW", "minWaiting");
		parametersMap.put( "maxW", "maxWaiting");
		parametersMap.put( "net", "networkOn"); 
		parametersMap.put( "mDate", "meanNumberDating"); 
		parametersMap.put( "sDate", "sdNumberDating"); 
		parametersMap.put( "wS", "weightedSwitch");
		parametersMap.put( "pTrue", "probRecognizeConcurrentTrue");
		parametersMap.put("pFalse", "probRecognizeConcurrentFalse");
		parametersMap.put("pC", "concurrentPenalty");
		parametersMap.put("rFC", "reportFileContents");
		parametersMap.put("mD", "meanDuration");
		parametersMap.put("sdD", "sdDuration");
		parametersMap.put("pIM", "probIncarcerationMale");
		parametersMap.put("pIP", "probIncarcerationPartnership");
		parametersMap.put("pIB", "probIncarcerationBreakup");
		parametersMap.put("iP", "incarcerationPenalty");
		parametersMap.put("qD", "qualityDistribution");
		parametersMap.put("df", "degFreedom");
		parametersMap.put("eQ", "eqQuality");
		parametersMap.put("tC", "toleranceChange");
		parametersMap.put("mT", "meanTime");
		parametersMap.put("sdT", "sdTime");
		parametersMap.put("pRc", "probRecidivate");}

	// Specify what appears in the repast parameter panel

	public String[] getInitParam() {
		String[] params = {"edgeListReportFileName", "numNodes", "worldXSize", "worldYSize",
				"updateEveryN", "LayoutType", "RemoveProb",
				   "StartRemoveAfter", "MaxDegree",  
				   "alphaMale", "lambdaMale", "alphaFemale", "lambdaFemale",
				   "meanQuality", "sdQuality", "meanAspiration", "sdAspiration",
				   "Rsub1", "Rsub0", "sexRatio","askMethod","minWaiting","maxWaiting",
				   "degreeDist", "Plot", "rDebug", "networkOn", "meanNumberDating", "sdNumberDating",
				   "weightedSwitch", "probRecognizeConcurrentTrue", "probRecognizeConcurrentFalse",
				   "concurrentPenalty", "reportFileContents", "meanDuration", "sdDuration", 
				   "probIncarcerationMale", "probIncarcerationPartnership", "probIncarcerationBreakup",
				   "incarcerationPenalty", "qualityDistribution", "degFreedom", "eqQuality", 
				   "toleranceChange", "meanTime", "sdTime", "probRecidivate"};
		return params;
	}

	/////////////////////////////////////////////////////////////////////////////
	// setters and getters 
	// These must be defined to be able to set parameters via GUI and run command 
	//
	// NB: some things can't be changed after run starts (eg sizeX,sizeY)

	public int getMaxDegree() { return maxDegree; }
	public void setMaxDegree(int degree) { maxDegree = degree;}

	public double getRsub0() { return Rsub0; }
	public void setRsub0(double R) { Rsub0 = R; }

	public double getRsub1() { return Rsub1; }
	public void setRsub1(double R) { Rsub1 = R; }

	public String getLayoutType() { return layoutType; }
	public void setLayoutType(String type) { layoutType = type; }

	public void setUpdateEveryN(int updates) { updateEveryN = updates; }
	public int getUpdateEveryN() { return updateEveryN; }

	public double getRemoveProb() { return removeProb; }
	public void setRemoveProb(double prob) { removeProb = prob; }

	public int getStartRemoveAfter() { return initialSteps; }
	public void setStartRemoveAfter(int steps) { initialSteps = steps; }

	public void setNumNodes(int n) { numNodes = n; }
	public int getNumNodes() { return numNodes; }

	public int getWorldXSize() { return worldXSize; }
    public void setWorldXSize(int size) {worldXSize = size; }
        
	public int getWorldYSize() { return worldYSize; }
	public void setWorldYSize(int size) { worldYSize = size; }

    public double getSexRatio() { return sexRatio; }
    public void setSexRatio( double sr ) { if (sr <= 1 && sr >=0) { sexRatio = sr; }}
	
	public double returnNumberCouples() { return (double) this.couples.size(); }
    public double returnNumberSexPartners() { return (double) this.married.size(); }
	
	public ArrayList<Couple> getCouples() { return couples; }
	public void setCouples(ArrayList<Couple> couples) { this.couples = couples; }

	public int getLastCoupleId() { return lastCoupleId; }
	public int getLastAgentId() { return lastAgentId; }	

	public boolean getDegreeDist() { return showHist; }
	public void setDegreeDist(boolean val) { showHist = val; }

	public boolean getPlot() { return showPlot; }
	public void setPlot(boolean val) { showPlot = val; }

    public int getRDebug() {return rDebug; }
    public void setRDebug(int r) { rDebug = r; }

    public double getMeanAspiration() { return meanAspiration; }
    public void setMeanAspiration( double a ) { meanAspiration = a; }

    public double getSdAspiration() { return sdAspiration;}
    public void setSdAspiration(double sda ) { sdAspiration = sda; }

    public double getMeanQuality() { return meanQuality; }
    public void setMeanQuality( double q ) { meanQuality = q; }

    public double getSdQuality() {return sdQuality; }
    public void setSdQuality( double sdq ) { sdQuality = sdq; }

    public double getMeanDuration() { return meanDuration; }
    public void setMeanDuration( double d ) { meanDuration = d; }

    public double getSdDuration() { return sdDuration; }
    public void setSdDuration( double sdd ) { sdDuration = sdd; }
	
    public int getMinWaiting() { return minWaiting;}
    public void setMinWaiting( int mw ) { minWaiting = mw; }
   
    public int getMaxWaiting() { return maxWaiting;}
    public void setMaxWaiting( int max ) { maxWaiting = max; }

    public double getAlphaMale() { return alphaMale; }
    public void setAlphaMale( double a ) { alphaMale = a; }

    public double getAlphaFemale() { return alphaFemale; }
    public void setAlphaFemale( double a ) { alphaFemale = a; }
        
    public double getLambdaMale() { return lambdaMale; }
    public void setLambdaMale( double l ) { lambdaMale = l; }
        
    public double getLambdaFemale() { return lambdaFemale; }
    public void setLambdaFemale( double l ) { lambdaFemale = l; }
        
    public double getProbRandomPartner() {return probRandomPartner; }
    public void setProbRandomPartner (double p ) { probRandomPartner = p; }
    
    public int getMeanNumberDating() { return meanNumberDating; }
    public void setMeanNumberDating( int m ) { meanNumberDating = m; }
    
    public int getSdNumberDating() { return sdNumberDating; }
    public void setSdNumberDating( int s ) { sdNumberDating = s; }
    
    public double getProbRecognizeConcurrentTrue() { return probRecognizeConcurrentTrue; }
    public void setProbRecognizeConcurrentTrue( double r ) { probRecognizeConcurrentTrue = r; }
    
    public double getProbRecognizeConcurrentFalse() { return probRecognizeConcurrentFalse; }
    public void setProbRecognizeConcurrentFalse( double r ) { probRecognizeConcurrentFalse = r; }
    
    public double getConcurrentPenalty() { return concurrentPenalty; }
    public void setConcurrentPenalty( double c ) { concurrentPenalty = c; }

    public boolean getShowPlot() { return showPlot; }
    public void setShowPlot( boolean p ) { showPlot = p; }
    
    public int getReportFileContents() { return reportFileContents; }
    public void setReportFileContents( int c ) { reportFileContents = c; }
    
    public double getProbIncarcerationMale() { return probIncarcerationMale; }
    public void setProbIncarcerationMale ( double i ) { probIncarcerationMale = i; }
    
    public double getProbIncarcerationPartnership() { return probIncarcerationPartnership; }
    public void setProbIncarcerationPartnership( double p ) { probIncarcerationPartnership = p; }
    
    public double getProbIncarcerationBreakup() { return probIncarcerationBreakup; }
    public void setProbIncarcerationBreakup( double b ) { probIncarcerationBreakup = b; }
    
    public double getIncarcerationPenalty() { return incarcerationPenalty; }
    public void setIncarcerationPenalty( double i ) { incarcerationPenalty = i; }
    
    public double getProbRecidivate() { return probRecidivate; }
    public void setProbRecidivate( double r ) { probRecidivate = r; }
        
    public void setAskMethod ( String s ) {
    	if ( s.equalsIgnoreCase( "men" ) )
    		ask = askingMethod.MENASK;
        else if (  s.equalsIgnoreCase( "women" ) )
            ask = askingMethod.WOMENASK;
        else if (  s.equalsIgnoreCase( "both" ) )
            ask = askingMethod.BOTHASK;
        else {
            System.out.printf( "illegal value for ask '%s'\n", s );
            return;
        }
        askMethod = s;
    }

    public String getAskMethod () {
    	switch ( ask ) {
        	case MENASK:
            	return "men";
            case WOMENASK:
            	 return "women";
            default:
             	 return "both";
    	}
    }

    public askingMethod getAsk () {
    	return ask;
    }
    
    public void setQualityDistribution ( String dist ) { qualityDistribution = dist; }
    public String getQualityDistribution () { return qualityDistribution; }

    public String getEdgeListReportFileName (){ return edgeListReportFileName; }
    public void setEdgeListReportFileName ( String n ) { edgeListReportFileName = n; }
    
    public boolean getNetworkOn(){ return networkOn; }
    public void setNetworkOn(boolean n){ networkOn = n; }
    
    public boolean getWeightedSwitch(){ return weightedSwitch; }
    public void setWeightedSwitch(boolean s){ weightedSwitch = s; }
    
    public boolean getEqQuality() { return eqQuality; }
    public void setEqQuality(boolean q){ eqQuality = q; }
    
    public int getDegFreedom() { return degFreedom; }
    public void setDegFreedom( int d ){ degFreedom = d; }
    
    public boolean getToleranceChange(){ return toleranceChange; }
    public void setToleranceChange( boolean change ){ toleranceChange = change; }
    
    public double getMeanTime() { return meanTime; }
    public void setMeanTime( double mean ) { meanTime = mean; }
    
    public double getSdTime() { return sdTime; }
    public void setSdTime( double sd ) { sdTime = sd; }

   	public static void main(String[] args) {
		uchicago.src.sim.engine.SimInit init =
			new uchicago.src.sim.engine.SimInit();
		Model model = new Model();
		init.loadModel(model, "", false);
	}

	///////////////////////////////////////////////////////////////////////////
	// userSetup()
	// called when user presses the "reset" button.
	// discard any existing model parts, and re-initialize as desired.
	//
	// NB: if you want values entered via the GUI to remain after restart,
	//     do not initialize them here.
	
	public void userSetup() {
		if ( rDebug > 0 )
			System.out.printf( "==> userSetup...\n" );

		agentList = new ArrayList();
		worldXSize = 500;
		worldYSize = 500;

		lastAgentId = 1;			// reset termite ID's to start at 0

	}

	///////////////////////////////////////////////////////////////////////////
	// userBuildModel
	// called when model initialized, eg, with Initialize button.
	// create all the agents
	// have some of them meet randomly
	// have them meet their neighbors and search for partners
	
	public void userBuildModel () {
		Vector vect = new Vector();
		vect.add("Fruch");
		vect.add("KK");
		vect.add("CircleLayout");
		ListPropertyDescriptor pd = new ListPropertyDescriptor("LayoutType", vect);
		descriptors.put("LayoutType", pd);

		if (rDebug > 1){
		    System.out.printf( "Time to start creating agents..." );
		}

		for (int n = 0; n < numNodes; n++) {
			int x = Random.uniform.nextIntFromTo(0, worldXSize - 1);
			int y = Random.uniform.nextIntFromTo(0, worldYSize - 1);
			Agent agent = new Agent(this, this.lastAgentId, x, y);
			initializeAgent(agent);	

			if (rDebug > 2){
			    System.out.printf( "==> Initialized agent %03d...\n", agent.getID() );
			    System.out.printf( "Gender: %s \n", agent.getGender());
			    System.out.printf( " Base Quality: %.3f \n", agent.getBaseQuality());
			    System.out.printf( "Aspiration Level: %.3f \n", agent.getAspirationLevel());
			    System.out.printf( "Maximum number of partners: %03d \n", agent.getMaxPartners());
			}
       				      
			agentList.add(agent);
			neverCoupled.add(agent);
			if (agent.getGender().compareTo("female") == 0) {
				females.add(agent);
			}
			else {
				males.add(agent);
			}			
			this.lastAgentId++;
		}
		if ( rDebug > 0 ){
			System.out.printf( "askMethod=%s, ask=%s\n", askMethod, ask );
		}
		
		Couple.setModel( this );
		
		if (networkOn == true){
			randomMeeting();
			neighborMeeting();
		}
		//chose partner search mechanism
	    switch ( ask ) {
	        case MENASK:
	        	malePartnerSearch();
	            break;
	        case WOMENASK:
	        	femalePartnerSearch();
	            break;
	        default:
	        	bothPartnerSearch();
	        }

	}

	///////////////////////////////////////////////////////////////////////////
	// initializeAgent ()
	// sets the initial levels for each of the agents
	// replaces: createAgentsAndAddToWorld
	// create numAgents agents, add to agentList and 
	// add to random locations in world
	//

	//RLR check if these are the distributions you want
        //based on the magic numbers from SJA

	public void initializeAgent(Agent agent) {
		double r;
		
		do 		r = Model.getNormalDouble( meanAspiration, sdAspiration );
		while ( r < 0 );
		agent.setAspirationLevel( (int) Math.round( r ) );
		
		do 		r = Model.getNormalDouble( 50, 25 );
		while ( r < 0 );
		agent.setAge( (int) Math.round( r ) );

		do 		r = Model.getNormalDouble( meanDuration, sdDuration );
		while ( r < 0 );
		agent.setCourtShipDuration( (int) Math.round( r ) );
		
		do 		r = Model.getNormalDouble( meanNumberDating, sdNumberDating );
		while ( r < 0 );
		agent.setMaxNumberDating( (int) Math.round( r ) );
		
		if (agent.getMaxNumberDating() < 1){
			agent.setMaxNumberDating(1);
		}

		agent.setWaitingThreshold( Model.getUniformIntFromTo( minWaiting, maxWaiting ));
		
		agent.setProbRecognizeConcurrentTrue( probRecognizeConcurrentTrue );
		agent.setProbRecognizeConcurrentFalse( probRecognizeConcurrentFalse );
		agent.setConcurrentPenalty( concurrentPenalty );
		
		agent.setProbIncarcerationPartnership( probIncarcerationPartnership );
		agent.setIncarcerationPenalty( incarcerationPenalty );

		if (  Model.getUniformDoubleFromTo( 0.0, 1.0 ) <= sexRatio) {
			agent.setGender("male");
		}
		else {
			agent.setGender("female");
		}
		
		if (agent.getGender() == "male"){
			do              r = Model.getGammaDouble( alphaMale, lambdaMale );
			while ( r < 0 );
			agent.setMaxPartners( (int) Math.round( r ) );
			if (agent.getMaxPartners() == 0){
				agent.setMaxPartners(1);
			}
		}
		if (agent.getGender() == "female"){
			do              r = Model.getGammaDouble( alphaFemale, lambdaFemale );
			while ( r < 0 );
			agent.setMaxPartners( (int) Math.round( r ) );
			if (agent.getMaxPartners() == 0){
				agent.setMaxPartners(1);
			}
		}
		if ( rDebug > 0 ){
			System.out.printf( "About to set quality");
		}
		setQuality( agent );
		
	}
	
	public void setQuality ( Agent agent ){
		double q;
		
		if ( eqQuality == true ){
			if ( qualityDistribution.equals( "normal" )){
				if ( rDebug > 1 ){
					System.out.printf( "distribution = normal");
				}
				do 		q = Model.getNormalDouble( meanQuality, sdQuality );
				while ( q < 0 );
				agent.setBaseQuality( (int) Math.round( q ) );
				if ( rDebug > 1 ){
					System.out.printf( "Agent Sex = %s, Quality Distribution = Normal \n", agent.getGender());
				}
			}
			else if ( qualityDistribution.equals( "chisquare" )) { 
				if ( rDebug > 1 ){
					System.out.printf( "distribution = %s \n", qualityDistribution);
				}
				do 		q = Model.getChiSquareDouble( degFreedom );
				while ( q < 0 );
				agent.setBaseQuality( (int) Math.round( q ) );
				if ( rDebug > 1 ){
					System.out.printf( "Agent Sex = %s, Quality Distribution = Chi Square \n", agent.getGender());
				}
			}
			else
				System.out.printf( "Something wrong with distribution assignment");
		}
		
		if ( eqQuality == false ){
			if ( agent.getGender() == "female" ){
				do 		q = Model.getNormalDouble( meanQuality, sdQuality );
				while ( q < 0 );
				agent.setBaseQuality( (int) Math.round( q ) );
				if ( rDebug > 1 ){
					System.out.printf( "Agent Sex = %s, Quality Distribution = Normal \n", agent.getGender());
				}
			}
			else if ( agent.getGender() == "male" ){
				do 		q = Model.getChiSquareDouble( degFreedom );
				while ( q < 0 );
				agent.setBaseQuality( (int) Math.round( q ) );
				if ( rDebug > 1 ){
					System.out.printf( "Agent Sex = %s, Quality Distribution = Chi Square \n", agent.getGender());
				}
			}
			else
				System.out.printf( "Something wrong with distribution assignment");
			}
	}


	///////////////////////////////////////////////////////////////////////////
	// step
	//
	// The top of the model's main dynamics.  This defines
	// what happens each time step (eg, user presses Step or Run buttons...).
	// Currently: 
	// - agents search for partners
	// - agents evaluate partners
	// - agents make proposals
	// - agents evaluate proposals
	// - couples are updated
	// - stepReport to write stats to the report file.
	//
	// NB: agents shuffled before being activated

	public void step () {
		if ( rDebug > 0 )
			System.out.printf( "==> Model step %.0f:\n", getTickCount() );

		for (Agent agent: agentList) { agent.update(); }
		if (networkOn == true){
			randomMeeting();
			neighborMeeting();
			removeRandomFriendship();
		}
		incarcerate();
		updateCouples();
		updateNeverCoupled();
		//probabilistic decision about random meetings versus network meetings
		if (  Model.getUniformDoubleFromTo( 0.0, 1.0 ) <= probRandomPartner) {
			switch ( ask ) {
        	case MENASK:
                randomMalePartnerSearch();
                //break;
        	case WOMENASK:
                randomFemalePartnerSearch();
                //break;
        	default:
                randomBothPartnerSearch();
	        }
		}
		else {
	        switch ( ask ) {
        	case MENASK:
                malePartnerSearch();
                //break;
        	case WOMENASK:
                femalePartnerSearch();
                //break;
        	default:
                bothPartnerSearch();
	        }
		}
		for (Agent agent: agentList) { agent.updatePartnerList(); }
		
		stepReport();		// write aggregate measures to report file

   		if ( rDebug > 2 ) {
			System.out.printf( "<== Model step done.\n" );
			System.out.printf( "Average Quality: %.3f \n", avgQuality);
			System.out.printf( "Average Aspiration Level: %.3f \n", avgAspiration);
			System.out.printf( "Average Number of Proposers: %.3f \n", avgNumProposers);
			System.out.printf( "Number of Couples: %.3f \n", returnNumberCouples());
			System.out.printf( "Number of Sex Partner Couples: %.3f \n", returnNumberSexPartners());
			System.out.printf( "Average Number of Partners: %.3f \n", avgNumPartners);
			System.out.printf( "Quality Correlation between Partners: %.3f \n", coupleCorr);
		    System.out.printf( "Average number of months dating: %.3f \n", avgMonthsDating);
		}

	}

        ///////////////////////////////////////////////////////////
        // incarceration 
        // - determines at each time step whether an agent is incarcerated
		// - determines the "incarcerationTime" for each agent

    public void incarcerate(){
	    for (Agent male: males) {
		    if ( male.getIncarcerated() == false ){
		    	if (Model.getUniformDoubleFromTo( 0.0, 1.0 ) <= probIncarcerationMale){
		    		male.setIncarcerated( true );
		    		male.setPrevIncarceration( true );
		    		male.decreaseAspiration();
		    		int time = (int) Math.round( Model.getNormalDouble( meanTime, sdTime ) );
		    		male.setIncarcerationTime(time);
		    		male.setIncarcerationStart( (int) getTickCount());
		    		if ( rDebug > 0 )
		    			System.out.printf( "==> Male %03d incarcerated %03d steps \n", male.getID(), male.getIncarcerationTime() );
		    	}
		    	if (male.getPrevIncarceration() == true){
		    		if (male.getIncarcerationStart() != getTickCount()){
		    			if (Model.getUniformDoubleFromTo( 0.0, 1.0) <= probRecidivate){
		    				male.setIncarcerated( true );
		    				male.decreaseAspiration();
		    				int time = (int) Math.round(Model.getNormalDouble( meanTime, sdTime ));
		    				male.setIncarcerationTime( time );
		    				male.setIncarcerationStart( (int) getTickCount());
		    				if (rDebug > 0 )
		    					System.out.printf( "==> Male %03d recidivates \n", male.getID() );
		    			}
		    		}
		    	}
		    	else
		    		male.setIncarcerated( false );
		    }
		    if (male.getIncarcerated() == true ){
		    	updateIncarceration( male );
		    }
	    }
	    
	    for (Agent female: females) {
	    	if ( female.getIncarcerated() == false ){
	    		if (Model.getUniformDoubleFromTo( 0.0, 1.0 ) <= probIncarcerationFemale){
	    			female.setIncarcerated( true );
	    			female.setPrevIncarceration( true );
	    			female.decreaseAspiration();
	    			int time = (int) Math.round( Model.getNormalDouble( meanTime, sdTime ) );
		    		female.setIncarcerationTime(time);
		    		female.setIncarcerationStart( (int) getTickCount());
	    			if ( rDebug > 0 ){
	    				System.out.printf( "==> Female %03d incarcerated %03d steps \n", female.getID(), female.getIncarcerationTime() );
	    		    }
	    		}
	    		else
	    			female.setIncarcerated( false );
	    	}	
	    	if ( female.getIncarcerated() == true ) {
	    		updateIncarceration( female );
	    	}
        }
    }

    ////////////////////////////////////////////////////////////////////////////////////////
    // updateIncarceration()
    // -based on "incarcerationTime" determines how long the individual will be listed
    //  as incarcerated
    // -unincarcerates when the time is up
    
    public void updateIncarceration( Agent agent ) {
    	int time = agent.getIncarcerationTime();
    	if ( time <= 0 ){
    		agent.setIncarcerated( false );
    		agent.setIncarcerationTime( 0 );
    		agent.setIncarcerationStart( 0 );
    		if ( rDebug > 1 ){
    			System.out.printf( "==> Agent %03d released \n", agent.getID());
    	    }
    	}
    	else {
    		time--;
    		agent.setIncarcerationTime( time );
        	if ( rDebug > 1 ){
    			System.out.printf( "==> Agent %03d has %03d steps remaining \n", agent.getID(), agent.getIncarcerationTime());
    	    }
    	}
    }
	
	/*
	 * "...At each time-step, we chose NpR0 pairs of vertices uniformly
	 * at random from the network to meet. If a pair meet who do not have
	 * a pre-existing connection, and if neither of them already has the
	 * maximum z* connections then a new connection is established
	 * between them...."
	 */

	private void randomMeeting() {
		int nPairs = (numNodes * (numNodes - 1)) / 2;
		int numPicks = (int) Math.round(nPairs * Rsub0);

		for (int n = 0; n < numPicks; n++) {
			int index = Random.uniform.nextIntFromTo(0, numNodes - 1);
			Agent iNode = (Agent) agentList.get(index);
			iNode.meetRandom(agentList, maxDegree);
		}
	}
	
	/**
	 * "...At each time-step, we choose NmR1 vertices at random, with
	 * probabilites proportional to Zi(Zi-1). For each vertex chosen we
	 * randomly choose on pair of its neighbots to meet, and establish a
	 * new connection between them if they do not have a preexisting
	 * connection and if neither of them already has the maximum number
	 * z* of connections...."
	 */
	
	private void neighborMeeting() {
		neighborMap.clear();
		double zeeSum = 0;
		for (int n = 0; n < numNodes; n++) {
			Agent node = (Agent) agentList.get(n);
			int degree = node.getOutDegree();
			if (degree > 1) {
				neighborMap.put(zeeSum, node);
				zeeSum += degree * (degree - 1);
			}
		}

		int nM = (int) (zeeSum / 2);   //will be even cause #edges even
		int numPicks = (int) Math.round(nM * Rsub1);

		for (int n = 0; n < numPicks; n++) {
			double pick = Random.uniform.nextDoubleFromTo(0, zeeSum);
			Agent node = (Agent) neighborMap.get(pick);

			if (node.getOutDegree() > 1) node.meetNeighbor(maxDegree);
		}
	}

	/*
	 * "...At each time-step, we choose NeY vertices at random, with
	 * probabilities proportional to Zi.  For each vertex chosen we
	 * choose one of its neighbors uniformly at random and delete the
	 * connection to that neighbor...."
	 */
	private void removeRandomFriendship() {

		removalMap.clear();
		double zeeSum = 0;
		int nEdges = 0;

		for (int i = 0; i < numNodes; i++) {
			Agent node = (Agent) agentList.get(i);
			int degree = node.getOutDegree();
			if (degree > 1) {
				removalMap.put(zeeSum, node);
				zeeSum += degree * (degree - 1);
			}
			nEdges += degree;
		}

		nEdges = nEdges / 2;
		int numPicks = (int) Math.round(nEdges * removeProb);

		for (int n = 0; n < numPicks; n++) {
			double pick = Random.uniform.nextDoubleFromTo(0, zeeSum);
			Agent node = (Agent) removalMap.get(pick);
			node.removeFriend();
		}
	}
	
        //updateCouples
        //
	public void updateCouples() {
	    for (Couple couple: couples) {

		    if (!couple.isMarried()) {
			couple.update();	
		    }
		    if (couple.isMarried() == true && married.contains(couple) == false) {
		    	married.add(couple); }
		}

		Iterator<Couple> coupleIterator = couples.iterator();
		while (coupleIterator.hasNext() ){
		    Couple couple = coupleIterator.next();
		    if (couple.isBroken() ){
			coupleIterator.remove();
				if (rDebug > 2 )
					System.out.printf ("\n Couple broke up \n");
		    }
		}
	}
	
	public void updateNeverCoupled() {
		Iterator<Agent> neverCoupledIterator = neverCoupled.iterator();
		while (neverCoupledIterator.hasNext() ) {
			Agent agent = neverCoupledIterator.next();
			if (agent.partners.size() > 0){
				neverCoupledIterator.remove();
				if ( rDebug > 2 ){
					System.out.printf("\n Agent %03d coupled \n", agent.getID() );
					System.out.printf("Size of neverCoupled = %03d", neverCoupled.size() );
					System.out.printf( "\n Agent %03d time to first partner = %03d \n", agent.getID(), agent.getTimeToFirstPartner() );
				}
			}		
			agent.increaseTimeToFirstPartner();
		}
	}

        //returnFemaleFriends
        //finds all out nodes who are female and adds them to a list
	public ArrayList<Agent> returnFemaleFriends(Agent male) {
		if (networkOn == true){
			friends = male.getOutNodes();
			for (Agent friend : friends) {
				if ((friend.getGender().compareTo("female") == 0)
					&& !femaleFriends.contains(friend)) {
						femaleFriends.add(friend);	
				}							
				ArrayList<Agent> secFriends = new ArrayList<Agent>(friend.getOutNodes());
				for (Agent secFriend : secFriends) {
					if ((secFriend.getGender().compareTo("female") == 0)
						&& !secFriends.contains(secFriend)) {
							femaleFriends.add(secFriend);	
					}
				}
			}
		}
		if (networkOn == false){
			if (getTickCount() ==0 ){
				for (double i = 0; i < maxDegree; i ++){
					int pick = Random.uniform.nextIntFromTo(0, (females.size()-1));
					femaleFriends.add(females.get(pick));
				}
			}
			else {
				int drop = Random.uniform.nextIntFromTo(0, femaleFriends.size()-1);
				femaleFriends.remove(drop);
				int pick = Random.uniform.nextIntFromTo(0, (females.size()-1));
				femaleFriends.add(females.get(pick));
			}
		}
		return femaleFriends;
	}
	
    //returnMaleFriends
    //finds all out nodes who are male and adds them to a list
public ArrayList<Agent> returnMaleFriends(Agent female) {
	if (networkOn == true){
		friends = female.getOutNodes();
		for (Agent friend : friends) {
			if ((friend.getGender().compareTo("male") == 0)
				&& !maleFriends.contains(friend)) {
					maleFriends.add(friend);	
			}							
			ArrayList<Agent> secFriends = new ArrayList<Agent>(friend.getOutNodes());
			for (Agent secFriend : secFriends) {
				if ((secFriend.getGender().compareTo("male") == 0)
					&& !secFriends.contains(secFriend)) {
					maleFriends.add(secFriend);	
				}
			}
		}
	}
	if (networkOn == false){
		if (getTickCount() ==0 ){
			for (double i = 0; i < maxDegree; i ++){
				int pick = Random.uniform.nextIntFromTo(0, (males.size()-1));
				maleFriends.add(males.get(pick));
			}
		}
		else {
			if (maleFriends.size() > 0){
				int drop = Random.uniform.nextIntFromTo(0, maleFriends.size()-1);
				maleFriends.remove(drop);
			}
			int pick = Random.uniform.nextIntFromTo(0, (males.size()-1));
			maleFriends.add(males.get(pick));
		}
	}
	return maleFriends;
}
	
   //malePartnerSearch()
   //looks for a partner among the list of female friends
   //sends proposal
   //proposal is evaluated

	public void malePartnerSearch() {
		SimUtilities.shuffle(this.agentList);
			
		for (Agent agent : this.agentList) {
			// agent.update();
			agent.getProposers().clear();  //clears previous list of proposers
		}
		
		SimUtilities.shuffle(this.males);
		SimUtilities.shuffle(this.females);
		
		for (Agent male : this.males) {

			innerLoop:for (Agent female : returnFemaleFriends(male)) {
				// loops through the list of female friends for the given male agent
			if (weightedSwitch == true){
				if (male.shouldSendOffer(female)) {
						// male determines whether he should send an offer to date
					female.getProposers().add(male);
					    // female adds male to list of proposers if he proposes
					if (rDebug > 2)
						System.out.printf( "Male proposes \n");
					   // break innerLoop; - this should just end when all
						//of the female friends have been addressed
				}		
			}
			if (weightedSwitch == false){
				if (male.shouldSendOfferUnweighted(female)) {
					// male determines whether he should send an offer to date
					female.getProposers().add(male);
				    // female adds male to list of proposers if he proposes
					if (rDebug > 2)
						System.out.printf( "Male proposes \n");
				   // break innerLoop; - this should just end when all
					//of the female friends have been addressed
				}
			}
			}
		}
		
		for (Agent female : this.females) {
			// this shuffles all of the female agents
			SimUtilities.shuffle(female.getProposers());			
			for (Agent proposer : female.getProposers()) {	
				// this loops through all of the proposers for each female
				if (weightedSwitch == true){
					if (female.receiveOffer(proposer)) {
						// she accepts the proposal based on whether or not the aspiration
						// level is lower than the quality (no mention of number of dates)
						Couple newCouple = makePartners(proposer, female);
						
						if (newCouple != null){
							this.couples.add(newCouple);
							if (rDebug > 2)
								System.out.printf( "Couple is now dating \n");
						}
					}	
				}
				if (weightedSwitch == false){
					if (female.receiveOfferUnweighted(proposer)) {
						// she accepts the proposal based on whether or not the aspiration
						// level is lower than the quality (no mention of number of dates)
						Couple newCouple = makePartners(proposer, female);
						if (newCouple != null){
							this.couples.add(newCouple);
							if (rDebug > 2)
								System.out.printf( "Couple is now dating \n");
						}
					}	
				}
			}						
		}				
	}
	
	//bothPartnerSearch()
	//agent looks for a partner among the list of 
	//male or female friends (depending on gender)
	//sends proposal
	//proposal is evaluated

	public void bothPartnerSearch() {
		SimUtilities.shuffle(this.agentList);
				
		for (Agent agent : this.agentList) {
			// agent.update();
			agent.getProposers().clear();  //clears previous list of proposers
		}
			
		SimUtilities.shuffle(this.males);
		SimUtilities.shuffle(this.females);
			
		for (Agent agent : this.agentList) {
			if (agent.getGender() == "female"){
				innerLoop:for (Agent male : returnMaleFriends(agent)) {
				// loops through the list of male friends for the given female agent
				if (weightedSwitch == true){
					if (agent.shouldSendOffer(male)) {
					// female determines whether he should send an offer to date
						male.getProposers().add(agent);
						// male adds male to list of proposers if he proposes
						if (rDebug > 2)
							System.out.printf( "Female proposes \n");
						   // break innerLoop; - this should just end when all
							//of the female friends have been addressed
					}		
				}
				if (weightedSwitch == false){
					if (agent.shouldSendOfferUnweighted(male)) {
						// female determines whether he should send an offer to date
							male.getProposers().add(agent);
							// male adds male to list of proposers if he proposes
							if (rDebug > 2)
								System.out.printf( "Female proposes \n");
							   // break innerLoop; - this should just end when all
								//of the female friends have been addressed
						}	
				}
				}
			}
			if (agent.getGender() == "male"){
				innerLoop:for (Agent female : returnFemaleFriends(agent)) {
				// loops through the list of female friends for the male agent
				if (weightedSwitch == true){
					if (agent.shouldSendOffer(female)) {
					// male determines whether he should send an offer to date
						female.getProposers().add(agent);
						// male adds male to list of proposers if he proposes
						if (rDebug > 2)
							System.out.printf( "Male proposes \n");
							// break innerLoop; - this should just end when all
							//of the female friends have been addressed
					}		
				}
				if (weightedSwitch == false){
					if (agent.shouldSendOfferUnweighted(female)) {
						// male determines whether he should send an offer to date
							female.getProposers().add(agent);
							// male adds male to list of proposers if he proposes
							if (rDebug > 2)
								System.out.printf( "Male proposes \n");
								// break innerLoop; - this should just end when all
								//of the female friends have been addressed
						}
				}
				}
			}
		}
			
		for (Agent agent : this.agentList) {
			// this shuffles all of the female agents
			SimUtilities.shuffle(agent.getProposers());			
			for (Agent proposer : agent.getProposers()) {	
			// this loops through all of the proposers for each female
				if (weightedSwitch == true){
					if (agent.receiveOffer(proposer)) {
						// agent accepts the proposal based on whether or not the aspiration
						// level is lower than the quality (no mention of number of dates)
						Couple newCouple = makePartners(proposer, agent);
						if (newCouple != null){
							this.couples.add(newCouple);
							if (rDebug > 2)
								System.out.printf( "Couple is now dating \n");
						}
					}		
				}
				if (weightedSwitch == false){
					if (agent.receiveOfferUnweighted(proposer)) {
						// agent accepts the proposal based on whether or not the aspiration
						// level is lower than the quality (no mention of number of dates)
						Couple newCouple = makePartners(proposer, agent);
						if (newCouple != null){
							this.couples.add(newCouple);
							if (rDebug > 2)
								System.out.printf( "Couple is now dating \n");
						}
					}		
				}
			}						
		}				
	}
		
		  //femalePartnerSearch()
		   //female agent looks for a partner among the list of male friends
		   //sends proposal
		   //proposal is evaluated

			public void femalePartnerSearch() {
				SimUtilities.shuffle(this.agentList);
					
				for (Agent agent : this.agentList) {
					// agent.update();
					agent.getProposers().clear();  //clears previous list of proposers
				}
				
				SimUtilities.shuffle(this.males);
				SimUtilities.shuffle(this.females);
				
				for (Agent female : this.females) {

					innerLoop:for (Agent male : returnMaleFriends(female)) {
						// loops through the list of male friends for the given female agent
					if (weightedSwitch == true){
						if (female.shouldSendOffer(male)) {
								// female determines whether he should send an offer to date
							male.getProposers().add(female);
							    // male adds male to list of proposers if he proposes
							if (rDebug > 2)
								System.out.printf( "Female proposes \n");
							   // break innerLoop; - this should just end when all
								//of the female friends have been addressed
						}		
					}
					if (weightedSwitch == false){
						if (female.shouldSendOfferUnweighted(male)) {
							// female determines whether he should send an offer to date
						male.getProposers().add(female);
						    // male adds male to list of proposers if he proposes
						if (rDebug > 2)
							System.out.printf( "Female proposes \n");
						   // break innerLoop; - this should just end when all
							//of the female friends have been addressed
					}	
					}
					}
				}
				
				for (Agent male : this.males) {
					// this shuffles all of the female agents
					SimUtilities.shuffle(male.getProposers());			
					for (Agent proposer : male.getProposers()) {	
						// this loops through all of the proposers for each female
						if (weightedSwitch == true){
							if (male.receiveOffer(proposer)) {
								// she accepts the proposal based on whether or not the aspiration
								// level is lower than the quality (no mention of number of dates)
								Couple newCouple = makePartners(proposer, male);
								if (newCouple != null){
									this.couples.add(newCouple);
									if (rDebug > 2)
										System.out.printf( "Couple is now dating \n");
								}
							}	
						}
						if (weightedSwitch == false){
							if (male.receiveOfferUnweighted(proposer)) {
								// she accepts the proposal based on whether or not the aspiration
								// level is lower than the quality (no mention of number of dates)
								Couple newCouple = makePartners(proposer, male);
								if (newCouple != null){
									this.couples.add(newCouple);
									if (rDebug > 2)
										System.out.printf( "Couple is now dating \n");
								}
							}	
						}
					}						
				}				
			}
	
	   //randomMalePartnerSearch()
	   //male agent looks for a partner among the list of all females
	   //sends proposal
	   //proposal is evaluated

		public void randomMalePartnerSearch() {
			SimUtilities.shuffle(this.agentList);
				
			for (Agent agent : this.agentList) {
				// agent.update();
				agent.getProposers().clear();  //clears previous list of proposers
			}
			
			SimUtilities.shuffle(this.males);
			SimUtilities.shuffle(this.females);
			
			for (Agent male : this.males) {
				Agent female = null;
				int index = Random.uniform.nextIntFromTo(0, females.size() - 1);
				female = (Agent) females.get(index);
				if (weightedSwitch == true){
					if (male.shouldSendOffer(female)) {
						// male determines whether he should send an offer to date
						female.getProposers().add(male);
						// female adds male to list of proposers if he proposes
						if (rDebug > 2){
							System.out.printf( "Male proposes \n");		
						}			
					}
				}
				if (weightedSwitch == false){
					if (male.shouldSendOfferUnweighted(female)) {
						// male determines whether he should send an offer to date
						female.getProposers().add(male);
						// female adds male to list of proposers if he proposes
						if (rDebug > 2){
							System.out.printf( "Male proposes \n");		
						}			
					}
				}
			}
			
			for (Agent female : this.females) {
				// this shuffles all of the female agents
				SimUtilities.shuffle(female.getProposers());			
				for (Agent proposer : female.getProposers()) {	
					// this loops through all of the proposers for each female
					if (weightedSwitch == true){
						if (female.receiveOffer(proposer)) {
							// she accepts the proposal based on whether or not the aspiration
							// level is lower than the quality (no mention of number of dates)
							Couple newCouple = makePartners(proposer, female);
							if (newCouple != null){
								this.couples.add(newCouple);
								if (rDebug > 2)
									System.out.printf( "Couple is now dating \n");
							}
						}
					}
					if (weightedSwitch == false){
						if (female.receiveOfferUnweighted(proposer)) {
							// she accepts the proposal based on whether or not the aspiration
							// level is lower than the quality (no mention of number of dates)
							Couple newCouple = makePartners(proposer, female);
							if (newCouple != null){
								this.couples.add(newCouple);
								if (rDebug > 2)
									System.out.printf( "Couple is now dating \n");
							}
						}
					}
				}						
			}				
		}

		   //randomFemalePartnerSearch()
		   //male agent looks for a partner among the list of all females
		   //sends proposal
		   //proposal is evaluated

			public void randomFemalePartnerSearch() {
				SimUtilities.shuffle(this.agentList);
					
				for (Agent agent : this.agentList) {
					// agent.update();
					agent.getProposers().clear();  //clears previous list of proposers
				}
				
				SimUtilities.shuffle(this.males);
				SimUtilities.shuffle(this.females);
				
				for (Agent female : this.females) {
					Agent male = null;
					int index = Random.uniform.nextIntFromTo(0, males.size() - 1);
					male = (Agent) males.get(index);
					if (weightedSwitch == true){
						if (female.shouldSendOffer(male)) {
							// female determines whether he should send an offer to date
							male.getProposers().add(female);
							// male adds male to list of proposers if he proposes
							if (rDebug > 2){
								System.out.printf( "Female proposes \n");		
							}			
						}
					}
					if (weightedSwitch == false){
						if (female.shouldSendOfferUnweighted(male)) {
							// female determines whether he should send an offer to date
							male.getProposers().add(female);
							// male adds male to list of proposers if he proposes
							if (rDebug > 2){
								System.out.printf( "Female proposes \n");		
							}			
						}
					}
				}
				
				for (Agent male : this.males) {
					// this shuffles all of the female agents
					SimUtilities.shuffle(male.getProposers());			
					for (Agent proposer : male.getProposers()) {	
						// this loops through all of the proposers for each female
						if (weightedSwitch == true){
							if (male.receiveOffer(proposer)) {
								Couple newCouple = makePartners(proposer, male);
								if (newCouple != null){
									this.couples.add(newCouple);
									if (rDebug > 2)
										System.out.printf( "Couple is now dating \n");
								}
							}
						}
						if (weightedSwitch == false){
							if (male.receiveOfferUnweighted(proposer)) {
								Couple newCouple = makePartners(proposer, male);
								if (newCouple != null){
									this.couples.add(newCouple);
									if (rDebug > 2)
										System.out.printf( "Couple is now dating \n");
								}
							}
						}
					}						
				}				
			}
			
	//randomBothPartnerSearch()
	//agent looks for a partner among the list of all females
	//sends proposal
	//proposal is evaluated

	public void randomBothPartnerSearch() {
		SimUtilities.shuffle(this.agentList);
						
		for (Agent agent : this.agentList) {
			// agent.update();
			agent.getProposers().clear();  //clears previous list of proposers
		}
					
		SimUtilities.shuffle(this.males);
		SimUtilities.shuffle(this.females);
		
		for (Agent agent : this.agentList) {
			if (agent.getGender() == "female"){
				Agent male = null;
				int index = Random.uniform.nextIntFromTo(0, males.size() - 1);
				male = (Agent) males.get(index);
				if (weightedSwitch == true){
					if (agent.shouldSendOffer(male)) {
						// male determines whether he should send an offer to date
						male.getProposers().add(agent);
						// male adds male to list of proposers if he proposes
						if (rDebug > 2){
							System.out.printf( "Female proposes \n");		
						}			
					}
				}
				if (weightedSwitch == false){
					if (agent.shouldSendOfferUnweighted(male)) {
						// male determines whether he should send an offer to date
						male.getProposers().add(agent);
						// male adds male to list of proposers if he proposes
						if (rDebug > 2){
							System.out.printf( "Female proposes \n");		
						}			
					}
				}
			}
			if (agent.getGender() == "male"){
				Agent female = null;
				int index = Random.uniform.nextIntFromTo(0, females.size() - 1);
				female = (Agent) females.get(index);
				if (weightedSwitch == true){
					if (agent.shouldSendOffer(female)) {
						// female determines whether he should send an offer to date
						female.getProposers().add(agent);
						// male adds male to list of proposers if he proposes
						if (rDebug > 2){
							System.out.printf( "Male proposes \n");		
						}			
					}
				}
				if (weightedSwitch == false){
					if (agent.shouldSendOfferUnweighted(female)) {
						// female determines whether he should send an offer to date
						female.getProposers().add(agent);
						// male adds male to list of proposers if he proposes
						if (rDebug > 2){
							System.out.printf( "Male proposes \n");		
						}			
					}
				}
			}
		}
		
		for (Agent agent : this.agentList) {
		// this shuffles all of the female agents
			SimUtilities.shuffle(agent.getProposers());			
			for (Agent proposer : agent.getProposers()) {	
			// this loops through all of the proposers for each female
				if (weightedSwitch == true){
					if (agent.receiveOffer(proposer)) {
						// agent accepts the proposal based on whether or not the aspiration
						// level is lower than the quality (no mention of number of dates)
						Couple newCouple = makePartners(proposer, agent);
						if (newCouple != null){
							this.couples.add(newCouple);
							if (rDebug > 2)
								System.out.printf( "Couple is now dating \n");
						}
					}
				}
				if (weightedSwitch == false){
					if (agent.receiveOfferUnweighted(proposer)) {
						// agent accepts the proposal based on whether or not the aspiration
						// level is lower than the quality (no mention of number of dates)
						Couple newCouple = makePartners(proposer, agent);
						if (newCouple != null){
							this.couples.add(newCouple);
							if (rDebug > 2)
								System.out.printf( "Couple is now dating \n");
						}
					}
				}
			}
		}
	}

        //makePartners
        //
	public Couple makePartners(Agent male, Agent female) {
		Couple couple = new Couple(this.lastCoupleId, male, 
				female);
		//adds the female to the list of individuals the male is dating
		if (male.potentialPartners.size() < male.getMaxNumberDating()){
			male.potentialPartners.add(female); 
		}
		if (male.potentialPartners.size() >= male.getMaxNumberDating()){
			if (rDebug > 1){
				System.out.printf("\n Male has too many dates (%3d) - must replace", male.potentialPartners.size());
			}
			Agent dropped = male.comparePartners(female, male.potentialPartners);
			if (dropped != null){
				male.courtshipBroken(dropped);
				dropped.courtshipBroken(male);
				for (Couple broken: male.currentCouples){
					if (broken.getMale() == male && broken.getFemale() == dropped){
						couple.setEndTime(getTickCount());
						couple.isBroken = true;
						male.currentCouples.remove(couple);
						female.currentCouples.remove(couple);
					}
				}
				male.potentialPartners.add(female); 
			}
			if (dropped == null && rDebug > 1){
				System.out.printf("Couldn't make partners - male has too many");
			}
		}

		//adds the male to the list of individuals the female is dating	
		if (female.potentialPartners.size() < female.getMaxNumberDating()){
			female.potentialPartners.add(male);
		}
		if (female.potentialPartners.size() >= female.getMaxNumberDating()){
			if (rDebug > 1){
				System.out.printf("\n Female has too many dates (%3d) - must replace", female.potentialPartners.size());
			}
			Agent dropped = female.comparePartners(male, female.potentialPartners);
			if (dropped != null){
				female.courtshipBroken(dropped);
				dropped.courtshipBroken(female);
				for (Couple broken: female.currentCouples){
					if (broken.getMale() == dropped && broken.getFemale() == female){
						couple.setEndTime(getTickCount());
						couple.isBroken = true;
						male.currentCouples.remove(couple);
						female.currentCouples.remove(couple);
					}
				}
				female.potentialPartners.add(male); 
			}
			if (dropped == null && rDebug > 1){
				System.out.printf("Couldn't make partners - female has too many");
			}
		}
		if (male.potentialPartners.contains(female) && female.potentialPartners.contains(male)){
			male.currentCouples.add(couple);
			female.currentCouples.add(couple);
			couple.setProbIncarcerationBreakup( probIncarcerationBreakup );
			this.lastCoupleId++;
			return couple;
		}
		else
			return null;
	}

	/////////////////////////////////////////////////////////////////////////////////
	// stepReport
	// called each model time step to write out lines that look like: 
        //     timeStep  ...data...data...data...
	// first it calls a method to calculate stats to be written.

	public void stepReport () {
		if ( rDebug > 0 )
			System.out.printf( "==> Model stepReport %.0f:\n", getTickCount() );

		calcStats();

		// set up a string with the values to write -- start with time step
		String s = String.format( "%5.0f  ", getTickCount() );

		// Append to String s here to write other data to report lines:

		if ( reportFileContents == 0 ){
			s += String.format( " %.3f ", percentSexuallyActiveF );
			s += String.format( " %.3f ", percentSexuallyActiveM );
			s += String.format( " %.3f ", fractionConcurrentM );
			s += String.format( " %.3f ", fractionConcurrentF );
			s += String.format( " %.3f ", avgPartnersInLastYearM );
			s += String.format( " %.3f ", avgPartnersInLastYearF );
			s += String.format( " %.3f ", avgLifetimePartnersM );
			s += String.format( " %.3f ", avgLifetimePartnersF );
			s += String.format( " %.3f ", coupleCorr );
			s += String.format( " %.3f ", avgMonthsDating );
		}
		
		if ( reportFileContents == 1){
			s += String.format( " %3d ", part1M );
			s += String.format( " %3d ", part2M );
			s += String.format( " %3d ", part3M );
			s += String.format( " %3d ", part4M );
			s += String.format( " %3d ", part5plusM);
			s += String.format( " %3d ", part1F );
			s += String.format( " %3d ", part2F );
			s += String.format( " %3d ", part3F );
			s += String.format( " %3d ", part4F );
			s += String.format( " %3d ", part5plusF);
			s += String.format( " %3d ", partLife1M );
			s += String.format( " %3d ", partLife2M );
			s += String.format( " %3d ", partLife3M );
			s += String.format( " %3d ", partLife4M );
			s += String.format( " %3d ", partLife5M );
			s += String.format( " %3d ", partLife6to10M );
			s += String.format( " %3d ", partLife11to20M );
			s += String.format( " %3d ", partLife21plusM );
			s += String.format( " %3d ", partLife1F );
			s += String.format( " %3d ", partLife2F );
			s += String.format( " %3d ", partLife3F );
			s += String.format( " %3d ", partLife4F );
			s += String.format( " %3d ", partLife5F );
			s += String.format( " %3d ", partLife6to10F );
			s += String.format( " %3d ", partLife11to20F );
			s += String.format( " %3d ", partLife21plusF );
			s += String.format( " %.3f ", percentSexuallyActiveM );
			s += String.format( " %.3f ", percentSexuallyActiveF );
		}
		
		if ( reportFileContents == 2 ){
			s += String.format( " %03d ", currentlyIncarceratedM );
			s += String.format( " %03d ", everIncarceratedM );
			s += String.format( " %.3f ", fractionConcurrentIM );
			s += String.format( " %.3f ", avgPartnersInLastYearIM );
			s += String.format( " %.3f ", avgLifetimePartnersIM );
		}		
		
		if ( reportFileContents == 3 ) {
			s += String.format( " %.3f ", avgQuality );
			s += String.format( " %.3f ", avgQualityMale );
			s += String.format( " %.3f ", avgQualityFemale );
		}

		// write it to the plain text report file, 'flush' buffer to file
		writeLineToPlaintextReportFile( s );
		getPlaintextReportFile().flush();

	}

	///////////////////////////////////////////////////////////////////////////////
	// calcStats
	// calculate various aggregrate measures, store in Model fields.
	
	public void calcStats () {
		calcAgentCharacteristics();
		calcNumPartners();
		calcRelationshipCharacteristics();
	}
	
	public void calcAgentCharacteristics(){
		double malesNeverCoupled = 0;
		for (Agent agent : neverCoupled ){
			if ( agent.getGender() == "male" ){
				malesNeverCoupled++;
			}
		}
		double femalesNeverCoupled = 0;
		for (Agent agent : neverCoupled ){
			if ( agent.getGender() == "female" ){
				femalesNeverCoupled++;
			}
		}
		
		double incMalesNeverCoupled = 0;
		for (Agent agent: neverCoupled){
			if (agent.getGender() == "male" && agent.getPrevIncarceration() == true ){
				incMalesNeverCoupled++;
			}
		}
		
	    //average quality over the run
        avgQuality = 0;
    	for ( Agent agent : agentList ) {
    		avgQuality += agent.getBaseQuality();
    	}

    	if ( agentList.size() > 1 ) {
			avgQuality /= agentList.size();
		}
    	
    	avgQualityFemale = 0;
    	for ( Agent agent: females ){
    		avgQualityFemale += agent.getBaseQuality();
    	}
    	if ( females.size() > 1 ) {
    		avgQualityFemale /= females.size();
    	}    	
    	
    	avgQualityMale = 0;
    	for ( Agent agent: males ){
    		avgQualityMale += agent.getBaseQuality();
    	}
    	if ( males.size() > 1 ) {
    		avgQualityMale /= males.size();
    	}
    	
    	//average aspiration level
		avgAspiration = 0;
		for ( Agent agent : agentList ) {
	    	avgAspiration += agent.getAspirationLevel();
		}

		if ( agentList.size() > 1 ) {
			avgAspiration /= agentList.size();
		}

     	//average number of proposers
		avgNumProposers = 0;
		for ( Agent agent : agentList ) {
			avgNumProposers += (agent.getProposers()).size();
		}
	
		if ( agentList.size() > 1 ) {
			avgNumProposers /= agentList.size();
		}
	     //calculate percentage of agents that are married
		percentMarried = 0;
		for ( Agent agent : agentList ) {
		    if(agent.partners.size() > 0) {
			percentMarried++;
		    }
		}

		if ( agentList.size() > 1 ) {
			percentMarried /= agentList.size();
		}
		
		percentSexuallyActiveF = 0;
		for ( Agent agent : females ) {
			if (agent.partners.size() > 0 ){
				percentSexuallyActiveF++;
			}
		}
		if ( females.size() > 1 ){
			percentSexuallyActiveF /= females.size();
		}
		
		percentSexuallyActiveM = 0;
		for ( Agent agent : males ){
			if (agent.partners.size() > 0 ){
				percentSexuallyActiveM++;
			}
		}
		if ( males.size() > 1 ){
			percentSexuallyActiveM /= males.size();
		}
		
		if ( rDebug > 0 ) {
			System.out.printf("\n %.3f Sexually Active Females", percentSexuallyActiveF);
			System.out.printf("\n %.3f Sexually Active Males", percentSexuallyActiveM);
		}
		
		// number of individuals currently and previously incarcerated
		currentlyIncarcerated = 0;
		for ( Agent agent: agentList ){
		    if (agent.getIncarcerated()) {
			currentlyIncarcerated++;
		    }
		}
		
		currentlyIncarceratedM = 0;
		for (Agent agent: males ){
			if (agent.getIncarcerated()) {
				currentlyIncarceratedM++;
			}
		}
		
		currentlyIncarceratedF = 0;
		for (Agent agent: females ){
			if (agent.getIncarcerated()) {
				currentlyIncarceratedF++;
			}
		}
		
		everIncarcerated = 0;
		for ( Agent agent: agentList ){
		    if (agent.getPrevIncarceration()){
			everIncarcerated++;
		    }
		}
		
		everIncarceratedF = 0;
		for ( Agent agent: females ){
		    if (agent.getPrevIncarceration()){
			everIncarceratedF++;
		    }
		}
		
		everIncarceratedM = 0;
		for ( Agent agent: males ){
		    if (agent.getPrevIncarceration()){
			everIncarceratedM++;
		    }
		}
		
		//average time to first partnership
		avgTimeToFirstPartner = 0;
		for (Agent agent: agentList) {
		    avgTimeToFirstPartner += agent.getTimeToFirstPartner();
		}
		for (Agent agent: neverCoupled) {
			avgTimeToFirstPartner -= agent.getTimeToFirstPartner();
		}
		
		if (agentList.size() > 1) {
		    avgTimeToFirstPartner /= agentList.size();
		}
		//% with concurrent partners at each timestep
		fractionConcurrent = 0;
		for ( Agent agent : agentList ) {
		    if(agent.partners.size() > 1) {
			fractionConcurrent++;
		    }
		}

		if ( agentList.size() > 1 ) {
			fractionConcurrent /= (agentList.size() - neverCoupled.size());
		}
		
		fractionConcurrentM = 0;
		for ( Agent agent : males ) {
			if(agent.partners.size() > 1) {
				fractionConcurrentM++;
			}
		}
		if ( males.size() > 1 ) {
			fractionConcurrentM /= (males.size() - malesNeverCoupled);
		}
		
		fractionConcurrentF = 0;
		for ( Agent agent : females ) {
		    if(agent.partners.size() > 1) {
			fractionConcurrentF++;
		    }
		}
		if ( females.size() > 1 ) {
			fractionConcurrentF /= (females.size() - femalesNeverCoupled);
		}
		
		fractionConcurrentIM = 0;
		for ( Agent agent: males ){
			if(agent.getPrevIncarceration() == true){
				if (agent.partners.size() > 1){
					fractionConcurrentIM++;
				}
			}
			if (everIncarcerated > 1) {
				fractionConcurrentIM /= everIncarcerated - incMalesNeverCoupled;
			}
		}

		
		//% with concurrent partners in the past year
		fractionConcurrentInYear = 0;
		for ( Agent agent : agentList ) {
		    if (agent.concurrentInPastYear == true){
		    	fractionConcurrentInYear++;
		    }
		}

		if ( agentList.size() > 1 ) {
			fractionConcurrentInYear /= (agentList.size() - neverCoupled.size());
		}
	}
	
	public void calcNumPartners(){
		double malesNeverCoupled = 0;
		for (Agent agent : neverCoupled ){
			if ( agent.getGender() == "male" ){
				malesNeverCoupled++;
			}
		}
		double femalesNeverCoupled = 0;
		for (Agent agent : neverCoupled ){
			if ( agent.getGender() == "female" ){
				femalesNeverCoupled++;
			}
		}
		
		double incMalesNeverCoupled = 0;
		for (Agent agent: neverCoupled){
			if (agent.getGender() == "male" && agent.getPrevIncarceration() == true ){
				incMalesNeverCoupled++;
			}
		}
	     //average number of partners
		avgNumPartners = 0;
		for ( Agent agent : agentList ) {
		    avgNumPartners += agent.partners.size();
		}
		if ( agentList.size() > 1 ) {
			avgNumPartners /= (agentList.size() - neverCoupled.size());
		}
		
		avgNumPartnersM = 0;
		for ( Agent agent : males ) {
		    avgNumPartnersM += agent.partners.size();
		}
		if ( males.size() > 1 ) {
			avgNumPartnersM /= (males.size() - malesNeverCoupled);
		}
		
		avgNumPartnersF = 0;
		for ( Agent agent : females ) {
		    avgNumPartnersF += agent.partners.size();
		}
		if ( females.size() > 1 ) {
			avgNumPartnersF /= (females.size() - femalesNeverCoupled);
		}
		
		//average number of dates
		avgNumPotential = 0;
		for (Agent agent: agentList) {
		    avgNumPotential += agent.potentialPartners.size();
		}
		if (agentList.size() > 1) {
		    avgNumPotential /= agentList.size();
		}
		//average number of lifetime partners for those individuals
		//who have had at least one partner (printed at each time step)
		avgLifetimePartners = 0;
		for ( Agent agent : agentList ) {
		    avgLifetimePartners += agent.lifetimePartners.size();
		}

		if ( agentList.size() > 1 ) {
			avgLifetimePartners /= (agentList.size() - neverCoupled.size());
		}
		
		avgLifetimePartnersM = 0;
		for ( Agent agent : males ) {
		    avgLifetimePartnersM += agent.lifetimePartners.size();
		}
		if ( males.size() > 1 ) {
			avgLifetimePartnersM /= (males.size() - malesNeverCoupled);
		}
		
		avgLifetimePartnersF = 0;
		for ( Agent agent : females ) {
		    avgLifetimePartnersF += agent.lifetimePartners.size();
		}
		if ( females.size() > 1 ) {
			avgLifetimePartnersF /= (females.size() - femalesNeverCoupled);
		}
		
		avgLifetimePartnersIM = 0;
		for ( Agent agent: males ){
			if (agent.getPrevIncarceration() == true){
				avgLifetimePartnersIM += agent.lifetimePartners.size();
			}
		}
		if ( everIncarcerated > 1 ){
			avgLifetimePartnersIM /= ( everIncarcerated - incMalesNeverCoupled);
		}
		
		//number of lifetime partners at each time step
		partLife1 = partLife2 = partLife3 = partLife4plus = 0;
		for ( Agent agent : agentList ) {
		    if (agent.lifetimePartners.size() == 1){
		    	partLife1++;
		    }
		    if (agent.lifetimePartners.size() == 2){
		    	partLife2++;
		    }
		    if (agent.lifetimePartners.size() == 3){
		    	partLife3++;
		    }
		    if (agent.lifetimePartners.size() >= 4){
		    	partLife4plus++;
		    }
		}
		
		//number of lifetime partners at each time step
		partLife1M = partLife2M = partLife3M = partLife4M = partLife5M = partLife6to10M = partLife11to20M = partLife21plusM = 0;
		for ( Agent agent : males ) {
		    if (agent.lifetimePartners.size() == 1){
		    	partLife1M++;
		    }
		    if (agent.lifetimePartners.size() == 2){
		    	partLife2M++;
		    }
		    if (agent.lifetimePartners.size() == 3){
		    	partLife3M++;
		    }
		    if (agent.lifetimePartners.size() == 4){
		    	partLife4M++;
		    }
		    if (agent.lifetimePartners.size() == 5){
		    	partLife5M++;
		    }
		    if (agent.lifetimePartners.size() >= 6 && agent.lifetimePartners.size() <= 10){
		    	partLife6to10M++;
		    }		    
		    if (agent.lifetimePartners.size() >= 11 && agent.lifetimePartners.size() <= 20){
		    	partLife11to20M++;
		    }
		    if (agent.lifetimePartners.size() >= 21){
		    	partLife21plusM++;
		    }
		}
		
		//number of lifetime partners at each time step
		partLife1F = partLife2F = partLife3F = partLife4F = partLife5F = partLife6to10F = partLife11to20F = partLife21plusF = 0;
		for ( Agent agent : females ) {
		    if (agent.lifetimePartners.size() == 1){
		    	partLife1F++;
		    }
		    if (agent.lifetimePartners.size() == 2){
		    	partLife2F++;
		    }
		    if (agent.lifetimePartners.size() == 3){
		    	partLife3F++;
		    }
		    if (agent.lifetimePartners.size() == 4){
		    	partLife4F++;
		    }
		    if (agent.lifetimePartners.size() == 5){
		    	partLife5F++;
		    }
		    if (agent.lifetimePartners.size() >= 6 && agent.lifetimePartners.size() <= 10){
		    	partLife6to10F++;
		    }		    
		    if (agent.lifetimePartners.size() >= 11 && agent.lifetimePartners.size() <= 20){
		    	partLife11to20F++;
		    }
		    if (agent.lifetimePartners.size() >= 21){
		    	partLife21plusF++;
		    }
		}
		//number of partners at each time step
		part1 = part2 = part3 = part4plus = 0;
		for ( Agent agent : agentList ) {
		    if (agent.partners.size() == 1){
		    	part1 ++;
		    }
		    if (agent.partners.size() == 2){
		    	part2 ++;
		    	agent.setEverConcurrent( true );
		    }
		    if (agent.partners.size() == 3){
		    	part3 ++;
		    	agent.setEverConcurrent( true );
		    }
		    if (agent.partners.size() >= 4){
		    	part4plus ++;
		    	agent.setEverConcurrent( true );
		    }
		}
		
		//number of partners in the past year
		if (getTickCount() < 52){
			avgPartnersInLastYear = 0;
		}
		else {
			avgPartnersInLastYear = 0;
			for ( Agent agent : agentList ) {
			    avgPartnersInLastYear += agent.partnersInLastYear.size();
			}

			if ( agentList.size() > 1 ) {
				avgPartnersInLastYear /= (agentList.size() - neverCoupled.size());
			}
		}
		
		if (getTickCount() < 52){
			avgPartnersInLastYearM = 0;
		}
		else {
			avgPartnersInLastYearM = 0;
			for ( Agent agent : males ) {
			    avgPartnersInLastYearM += agent.partnersInLastYear.size();
			}

			if ( males.size() > 1 ) {
				avgPartnersInLastYearM /= (males.size() - malesNeverCoupled);
			}
		}
		
		if (getTickCount() < 52){
			avgPartnersInLastYearF = 0;
		}
		else {
			avgPartnersInLastYearF = 0;
			for ( Agent agent : females ) {
			    avgPartnersInLastYearF += agent.partnersInLastYear.size();
			}

			if ( females.size() > 1 ) {
				avgPartnersInLastYearF /= (females.size() - femalesNeverCoupled);
			}
		}
		
		//number of partners in the past year
		part1M = part2M = part3M = part4M = part5plusM = 0;
		if (getTickCount() > 52){
		for ( Agent agent : males ) {
		    if (agent.partnersInLastYear.size() == 1){
		    	part1M ++;
		    }
		    if (agent.partnersInLastYear.size() == 2){
		    	part2M ++;
		    }
		    if (agent.partnersInLastYear.size() == 3){
		    	part3M ++;
		    }
		    if (agent.partnersInLastYear.size() == 4){
		    	part4M ++;
		    }
		    if (agent.partnersInLastYear.size() >= 5){
		    	part5plusM ++;
		    }
		}
		}
		
		//number of partners in the past year
		part1F = part2F = part3F = part4F = part5plusF = 0;
		if (getTickCount() > 52){
		for ( Agent agent : females ) {
		    if (agent.partnersInLastYear.size() == 1){
		    	part1F ++;
		    }
		    if (agent.partnersInLastYear.size() == 2){
		    	part2F ++;
		    }
		    if (agent.partnersInLastYear.size() == 3){
		    	part3F ++;
		    }
		    if (agent.partnersInLastYear.size() == 4){
		    	part4F ++;
		    }
		    if (agent.partnersInLastYear.size() >= 5){
		    	part5plusF ++;
		    }
		}
		}
		
		if (getTickCount() < 52){
			avgPartnersInLastYearIM = 0;
		}
		else {
			avgPartnersInLastYearIM = 0;
			for ( Agent agent : males ) {
				if (agent.getPrevIncarceration() == true ){
					avgPartnersInLastYearIM += agent.partnersInLastYear.size();
				}
			}

			if ( everIncarcerated > 1 ) {
				avgPartnersInLastYearIM /= (everIncarcerated - incMalesNeverCoupled);
			}
		}
		
	}
	
	public void calcRelationshipCharacteristics(){
	     //correlation between the quality measures for each couple
		coupleCorr = 0;

		if(married.size() > 3) {

		    double[] males = new double[married.size()];
		    double[] females = new double[married.size()];
		
		    for (int i = 0; i < married.size(); i++) {
		
		    	Agent male = married.get(i).getMale();
		    	Agent female = married.get(i).getFemale();
    
		    	males[i] = male.getBaseQuality();
		    	females[i] = female.getBaseQuality();
		    }
		    if (rDebug > 3){
		    	for (int i = 0; i < married.size(); i++) {
		    		System.out.printf("male quality");
		    		System.out.println(males[i]);
		    		System.out.printf("female quality");
		    		System.out.println(females[i]);		    		
		    	}

		    }
			
		    PairedData quality = new PairedData(males, females);

			LinearCorrelation corr = new LinearCorrelation(quality);

			coupleCorr = corr.getR();
		 }

	     //calculate the average duration of partnerships
		avgMonthsDating=0;
		for (Couple couple: couples ){
		    avgMonthsDating += couple.getNumMonthsDating();
		}
		    
		if (couples.size() > 1) {
		    avgMonthsDating /= couples.size();
		}
		
	}

	///////////////////////////////////////////////////////////////////////////////
	// writeHeaderCommentsToReportFile
	// customize to match what you are writing to the report files in stepReport.
	
	public void writeHeaderCommentsToReportFile () {
		
		if ( reportFileContents == 0 ){
		writeLineToPlaintextReportFile( "#                    " );
		writeLineToPlaintextReportFile( "#       %      %      %M     %F      Year   Year  avg    avg    Couple  avg" );
		writeLineToPlaintextReportFile( "# time  SAF    SAM    Conc   Conc    M      F     lifeM  lifeF  Corr    Duration" );
		}
		
		if ( reportFileContents == 1){
			writeLineToPlaintextReportFile( "#                    " );
			writeLineToPlaintextReportFile( "#       Year Partners - M     Year Partners - F       Lifetime Partners - M                       Lifetime Partners - F                       %   %" );
			writeLineToPlaintextReportFile( "# time  1    2    3    4+     1    2    3    4    5+  1    2    3    4    5   6-10  11-20  21+   1    2    3    4    5   6-10  11-20  21+   SAM SAF" );
		}	
		
		if ( reportFileContents == 2){
			writeLineToPlaintextReportFile( "#                    " );
			writeLineToPlaintextReportFile( "#       incM incM  Conc IM     IM" );
			writeLineToPlaintextReportFile( "# time  curr ever  %IM  Year   Life" );
		}	
		
		if ( reportFileContents == 3){
			writeLineToPlaintextReportFile( "#                    " );
			writeLineToPlaintextReportFile( "#       avg avg  avg " );
			writeLineToPlaintextReportFile( "# time  Q   QM   QF  " );
		}	
	}
	
	///////////////////////////////////////////////////////////////////////////////
	// writeEdgeList
	// creates the edgelist with start and stop times and prints this to a report file
	// -currently is only called while the end of the run is being processed
	
	public void writeEdgeList() {
		edgeListReportFileName = "edgeList";
		//edgeListReportFile = startReportFile( edgeListReportFileName );
		edgeListPlainTextReportFile = startPlainTextReportFile ( edgeListReportFileName );
		if (rDebug > 2){
			System.out.printf("edgeList Report Created \n");
		}
		String e; 
		for (Couple couple : married){
			if (couple.getEndTime() == 999){
				couple.setEndTime(getTickCount());
			}
			Agent male = couple.getMale();
			Agent female = couple.getFemale();
			e = String.format("\n %03d ", male.getID() );
			e += String.format(" %03d ", female.getID() );
			e += String.format(" %5.0f ", couple.getStartTime() );
			e += String.format(" %5.0f \n", couple.getEndTime() );
			// write the line to the xml report file
			// writeLineToReportFile( e, edgeListReportFile );		
			// write it to the plain text report file, 'flush' buffer to file
			writeLineToReportFile( e, edgeListPlainTextReportFile );
			getPlaintextReportFile().flush();
		}
		// endReportFile( edgeListReportFile );
		endPlainTextReportFile( edgeListPlainTextReportFile );
	}	

	//////////////////////////////////////////////////////////////////////////////////
	// printProjectHelp
	// this should be filled in with some help to get from running as
	//        bin/guirun.sh -h
	//
	
	public void printProjectHelp() {
		System.out.printf( "\n%s -- Incarcertation \n", getName() );

		System.out.printf( "\n" );

		printParametersMap();

		System.exit( 0 );

	}



	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	//
	//            USUALLY NO NEED TO CHANGE THINGS BELOW HERE
	//
	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////



	////////////////////////////////////////////////////////////////////////////
	// constructor, if need to do anything special.
	public Model () {

	}

	///////////////////////////////////////////////////////////////////////////
	// setup
	// set generic defaults after a run start or restart
	// calls userSetup() which the model-author defines
	//   for the specific model.

	public void setup() {
		schedule = null;
		if ( rDebug > 1 )
			System.out.printf( "==> Model-setup...\n" );

		userSetup();

		System.gc ();   // garabage collection of discarded objects
		super.setup();  // THIS SHOULD BE CALLED after setting defaults in setup().
		schedule = new Schedule (1);  // create AFTER calling super.setup()

		if ( rDebug > 1 )
			System.out.printf( "\n<=== Model-setup() done.\n" );

	}


	///////////////////////////////////////////////////////////////////////////
	// buildModel
	// build the generic "architecture" for the model,
	// and call userBuildModel() which the model-author defines
	// to create the model-specific components.

	public void buildModel () {
		if ( rDebug > 1 )
			System.out.printf( "==> buildModel...\n" );

		// CALL FIRST -- defined in super class -- it starts RNG, etc
		buildModelStart();

		userBuildModel();

		// some post-load finishing touches
		startReportFile();

		// you probably don't want to remove any of the following
		// calls to process parameter changes and write the
		// initial state to the report file.
		// NB -> you might remove/add more agentChange processing
        applyAnyStoredChanges();
        stepReport();
        // getReportFile().flush();
        getPlaintextReportFile().flush();

		if ( rDebug > 1 )
			System.out.printf( "<== buildModel done.\n" );
	}

	//////////////////////////////////////////////////////////////////////////////
	//  unclear what these are for, but I don't want to cut them out
	
	public Schedule getSchedule () {	return schedule; }

	public String getName () { return "Model"; }



	/////////////////////////////////////////////////////////////////////////////
	// processEndOfRun
	// called once, at end of run.
	// writes some final info, closes report files, etc.
	public void processEndOfRun ( ) {

		if ( rDebug > 0 )  
			System.out.printf("\n\n===== processEndOfRun =====\n\n" );
		applyAnyStoredChanges();
		endReportFile();
		writeEdgeList();


		this.fireStopSim();
	}

}